@@ -22,11 +22,11 @@
 #define BFS_H
 
 #ifndef BFS_VERSION
-#	define BFS_VERSION "1.7"
+#	define BFS_VERSION "2.1"
 #endif
 
 #ifndef BFS_HOMEPAGE
-#	define BFS_HOMEPAGE "https://github.com/tavianator/bfs"
+#	define BFS_HOMEPAGE "https://tavianator.com/projects/bfs.html"
 #endif
 
 #endif // BFS_H
