@@ -73,7 +73,7 @@ struct bftw_file {
 	int fd;
 
 	/** This file's type, if known. */
-	enum bftw_typeflag typeflag;
+	enum bftw_type type;
 	/** The device number, for cycle detection. */
 	dev_t dev;
 	/** The inode number, for cycle detection. */
@@ -300,7 +300,7 @@ static size_t bftw_child_nameoff(const struct bftw_file *parent) {
 /** Create a new bftw_file. */
 static struct bftw_file *bftw_file_new(struct bftw_cache *cache, struct bftw_file *parent, const char *name) {
 	size_t namelen = strlen(name);
-	size_t size = sizeof(struct bftw_file) + namelen + 1;
+	size_t size = BFS_FLEX_SIZEOF(struct bftw_file, name, namelen + 1);
 
 	struct bftw_file *file = malloc(size);
 	if (!file) {
@@ -325,7 +325,7 @@ static struct bftw_file *bftw_file_new(struct bftw_cache *cache, struct bftw_fil
 	file->refcount = 1;
 	file->fd = -1;
 
-	file->typeflag = BFTW_UNKNOWN;
+	file->type = BFTW_UNKNOWN;
 	file->dev = -1;
 	file->ino = -1;
 
@@ -426,7 +426,7 @@ static int bftw_file_open(struct bftw_cache *cache, struct bftw_file *file, cons
 	// Handle ENAMETOOLONG by manually traversing the path component-by-component
 
 	// -1 to include the root, which has depth == 0
-	size_t offset = base ? base->depth : -1;
+	size_t offset = base ? base->depth : (size_t)-1;
 	size_t levels = file->depth - offset;
 	if (levels < 2) {
 		return fd;
@@ -754,7 +754,7 @@ err:
 	return -1;
 }
 
-enum bftw_typeflag bftw_mode_typeflag(mode_t mode) {
+enum bftw_type bftw_mode_to_type(mode_t mode) {
 	switch (mode & S_IFMT) {
 #ifdef S_IFBLK
 	case S_IFBLK:
@@ -802,7 +802,7 @@ enum bftw_typeflag bftw_mode_typeflag(mode_t mode) {
 	}
 }
 
-static enum bftw_typeflag bftw_dirent_typeflag(const struct dirent *de) {
+static enum bftw_type bftw_dirent_type(const struct dirent *de) {
 #if defined(_DIRENT_HAVE_D_TYPE) || defined(DT_UNKNOWN)
 	switch (de->d_type) {
 #ifdef DT_BLK
@@ -852,7 +852,7 @@ static enum bftw_typeflag bftw_dirent_typeflag(const struct dirent *de) {
 }
 
 /** Cached bfs_stat(). */
-static const struct bfs_stat *bftw_stat_impl(struct BFTW *ftwbuf, struct bftw_stat *cache, enum bfs_stat_flag flags) {
+static const struct bfs_stat *bftw_stat_impl(struct BFTW *ftwbuf, struct bftw_stat *cache, enum bfs_stat_flags flags) {
 	if (!cache->buf) {
 		if (cache->error) {
 			errno = cache->error;
@@ -866,7 +866,7 @@ static const struct bfs_stat *bftw_stat_impl(struct BFTW *ftwbuf, struct bftw_st
 	return cache->buf;
 }
 
-const struct bfs_stat *bftw_stat(const struct BFTW *ftwbuf, enum bfs_stat_flag flags) {
+const struct bfs_stat *bftw_stat(const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
 	struct BFTW *mutbuf = (struct BFTW *)ftwbuf;
 	const struct bfs_stat *ret;
 
@@ -886,18 +886,30 @@ const struct bfs_stat *bftw_stat(const struct BFTW *ftwbuf, enum bfs_stat_flag f
 	return ret;
 }
 
-enum bftw_typeflag bftw_typeflag(const struct BFTW *ftwbuf, enum bfs_stat_flag flags) {
+const struct bfs_stat *bftw_cached_stat(const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
+	if (flags & BFS_STAT_NOFOLLOW) {
+		return ftwbuf->lstat_cache.buf;
+	} else if (ftwbuf->stat_cache.buf) {
+		return ftwbuf->stat_cache.buf;
+	} else if ((flags & BFS_STAT_TRYFOLLOW) && is_nonexistence_error(ftwbuf->stat_cache.error)) {
+		return ftwbuf->lstat_cache.buf;
+	} else {
+		return NULL;
+	}
+}
+
+enum bftw_type bftw_type(const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
 	if (ftwbuf->stat_flags & BFS_STAT_NOFOLLOW) {
-		if ((flags & BFS_STAT_NOFOLLOW) || ftwbuf->typeflag != BFTW_LNK) {
-			return ftwbuf->typeflag;
+		if ((flags & BFS_STAT_NOFOLLOW) || ftwbuf->type != BFTW_LNK) {
+			return ftwbuf->type;
 		}
-	} else if ((flags & (BFS_STAT_NOFOLLOW | BFS_STAT_TRYFOLLOW)) == BFS_STAT_TRYFOLLOW || ftwbuf->typeflag == BFTW_LNK) {
-		return ftwbuf->typeflag;
+	} else if ((flags & (BFS_STAT_NOFOLLOW | BFS_STAT_TRYFOLLOW)) == BFS_STAT_TRYFOLLOW || ftwbuf->type == BFTW_LNK) {
+		return ftwbuf->type;
 	}
 
 	const struct bfs_stat *statbuf = bftw_stat(ftwbuf, flags);
 	if (statbuf) {
-		return bftw_mode_typeflag(statbuf->mode);
+		return bftw_mode_to_type(statbuf->mode);
 	} else {
 		return BFTW_ERROR;
 	}
@@ -934,16 +946,16 @@ static bool bftw_need_stat(const struct bftw_state *state) {
 	}
 
 	const struct BFTW *ftwbuf = &state->ftwbuf;
-	if (ftwbuf->typeflag == BFTW_UNKNOWN) {
+	if (ftwbuf->type == BFTW_UNKNOWN) {
 		return true;
 	}
 
-	if (ftwbuf->typeflag == BFTW_LNK && !(ftwbuf->stat_flags & BFS_STAT_NOFOLLOW)) {
+	if (ftwbuf->type == BFTW_LNK && !(ftwbuf->stat_flags & BFS_STAT_NOFOLLOW)) {
 		return true;
 	}
 
-	if (ftwbuf->typeflag == BFTW_DIR) {
-		if (state->flags & (BFTW_DETECT_CYCLES | BFTW_MOUNT | BFTW_XDEV)) {
+	if (ftwbuf->type == BFTW_DIR) {
+		if (state->flags & (BFTW_DETECT_CYCLES | BFTW_SKIP_MOUNTS | BFTW_PRUNE_MOUNTS)) {
 			return true;
 		}
 #if __linux__
@@ -1007,7 +1019,7 @@ static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 	ftwbuf->root = file ? file->root->name : ftwbuf->path;
 	ftwbuf->depth = 0;
 	ftwbuf->visit = visit;
-	ftwbuf->typeflag = BFTW_UNKNOWN;
+	ftwbuf->type = BFTW_UNKNOWN;
 	ftwbuf->error = reader->error;
 	ftwbuf->at_fd = AT_FDCWD;
 	ftwbuf->at_path = ftwbuf->path;
@@ -1019,12 +1031,12 @@ static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 	if (de) {
 		parent = file;
 		ftwbuf->depth = file->depth + 1;
-		ftwbuf->typeflag = bftw_dirent_typeflag(de);
+		ftwbuf->type = bftw_dirent_type(de);
 		ftwbuf->nameoff = bftw_child_nameoff(file);
 	} else if (file) {
 		parent = file->parent;
 		ftwbuf->depth = file->depth;
-		ftwbuf->typeflag = file->typeflag;
+		ftwbuf->type = file->type;
 		ftwbuf->nameoff = file->nameoff;
 	}
 
@@ -1044,13 +1056,13 @@ static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 	}
 
 	if (ftwbuf->error != 0) {
-		ftwbuf->typeflag = BFTW_ERROR;
+		ftwbuf->type = BFTW_ERROR;
 		return;
 	}
 
-	int follow_flags = BFTW_LOGICAL;
+	int follow_flags = BFTW_FOLLOW_ALL;
 	if (ftwbuf->depth == 0) {
-		follow_flags |= BFTW_COMFOLLOW;
+		follow_flags |= BFTW_FOLLOW_ROOTS;
 	}
 	bool follow = state->flags & follow_flags;
 	if (follow) {
@@ -1061,18 +1073,18 @@ static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 	if (bftw_need_stat(state)) {
 		statbuf = bftw_stat(ftwbuf, ftwbuf->stat_flags);
 		if (statbuf) {
-			ftwbuf->typeflag = bftw_mode_typeflag(statbuf->mode);
+			ftwbuf->type = bftw_mode_to_type(statbuf->mode);
 		} else {
-			ftwbuf->typeflag = BFTW_ERROR;
+			ftwbuf->type = BFTW_ERROR;
 			ftwbuf->error = errno;
 			return;
 		}
 	}
 
-	if (ftwbuf->typeflag == BFTW_DIR && (state->flags & BFTW_DETECT_CYCLES)) {
+	if (ftwbuf->type == BFTW_DIR && (state->flags & BFTW_DETECT_CYCLES)) {
 		for (const struct bftw_file *ancestor = parent; ancestor; ancestor = ancestor->parent) {
 			if (ancestor->dev == statbuf->dev && ancestor->ino == statbuf->ino) {
-				ftwbuf->typeflag = BFTW_ERROR;
+				ftwbuf->type = BFTW_ERROR;
 				ftwbuf->error = ELOOP;
 				return;
 			}
@@ -1122,12 +1134,12 @@ static enum bftw_action bftw_visit(struct bftw_state *state, const char *name, e
 	bftw_init_ftwbuf(state, visit);
 
 	// Never give the callback BFTW_ERROR unless BFTW_RECOVER is specified
-	if (ftwbuf->typeflag == BFTW_ERROR && !(state->flags & BFTW_RECOVER)) {
+	if (ftwbuf->type == BFTW_ERROR && !(state->flags & BFTW_RECOVER)) {
 		state->error = ftwbuf->error;
 		return BFTW_STOP;
 	}
 
-	if ((state->flags & BFTW_MOUNT) && bftw_is_mount(state, name)) {
+	if ((state->flags & BFTW_SKIP_MOUNTS) && bftw_is_mount(state, name)) {
 		return BFTW_PRUNE;
 	}
 
@@ -1143,12 +1155,12 @@ static enum bftw_action bftw_visit(struct bftw_state *state, const char *name, e
 		return BFTW_STOP;
 	}
 
-	if (visit != BFTW_PRE || ftwbuf->typeflag != BFTW_DIR) {
+	if (visit != BFTW_PRE || ftwbuf->type != BFTW_DIR) {
 		ret = BFTW_PRUNE;
 		goto done;
 	}
 
-	if ((state->flags & BFTW_XDEV) && bftw_is_mount(state, name)) {
+	if ((state->flags & BFTW_PRUNE_MOUNTS) && bftw_is_mount(state, name)) {
 		ret = BFTW_PRUNE;
 		goto done;
 	}
@@ -1174,7 +1186,7 @@ static int bftw_push(struct bftw_state *state, const char *name, bool fill_id) {
 
 	struct dirent *de = state->reader.de;
 	if (de) {
-		file->typeflag = bftw_dirent_typeflag(de);
+		file->type = bftw_dirent_type(de);
 	}
 
 	if (fill_id) {
@@ -1288,7 +1300,7 @@ static enum bftw_action bftw_release_reader(struct bftw_state *state, enum bftw_
 static enum bftw_action bftw_release_file(struct bftw_state *state, enum bftw_release_flags flags) {
 	enum bftw_action ret = BFTW_CONTINUE;
 
-	if (!(state->flags & BFTW_DEPTH)) {
+	if (!(state->flags & BFTW_POST_ORDER)) {
 		flags = 0;
 	}
 	bool visit = flags & BFTW_VISIT_FILE;
@@ -1346,6 +1358,21 @@ static int bftw_state_destroy(struct bftw_state *state) {
 	return state->error ? -1 : 0;
 }
 
+/** Start a batch of files. */
+static void bftw_batch_start(struct bftw_state *state) {
+	if (state->strategy == BFTW_DFS) {
+		state->queue.target = &state->queue.head;
+	}
+	state->batch = state->queue.target;
+}
+
+/** Finish adding a batch of files. */
+static void bftw_batch_finish(struct bftw_state *state) {
+	if (state->flags & BFTW_SORT) {
+		state->queue.target = bftw_sort_files(state->batch, state->queue.target);
+	}
+}
+
 /**
  * Streaming mode: visit files as they are encountered.
  */
@@ -1355,6 +1382,9 @@ static int bftw_stream(const struct bftw_args *args) {
 		return -1;
 	}
 
+	assert(!(state.flags & BFTW_SORT));
+
+	bftw_batch_start(&state);
 	for (size_t i = 0; i < args->npaths; ++i) {
 		const char *path = args->paths[i];
 
@@ -1371,10 +1401,12 @@ static int bftw_stream(const struct bftw_args *args) {
 			goto done;
 		}
 	}
+	bftw_batch_finish(&state);
 
 	while (bftw_pop(&state) > 0) {
 		struct bftw_reader *reader = bftw_open(&state);
 
+		bftw_batch_start(&state);
 		while (bftw_reader_read(reader) > 0) {
 			const char *name = reader->de->d_name;
 
@@ -1391,6 +1423,7 @@ static int bftw_stream(const struct bftw_args *args) {
 				goto done;
 			}
 		}
+		bftw_batch_finish(&state);
 
 		if (bftw_release_reader(&state, BFTW_VISIT_ALL) == BFTW_STOP) {
 			goto done;
@@ -1404,21 +1437,6 @@ done:
 	return bftw_state_destroy(&state);
 }
 
-/** Start a batch of files. */
-static void bftw_batch_start(struct bftw_state *state) {
-	if (state->strategy == BFTW_DFS) {
-		state->queue.target = &state->queue.head;
-	}
-	state->batch = state->queue.target;
-}
-
-/** Finish adding a batch of files. */
-static void bftw_batch_finish(struct bftw_state *state) {
-	if (state->flags & BFTW_SORT) {
-		state->queue.target = bftw_sort_files(state->batch, state->queue.target);
-	}
-}
-
 /**
  * Batching mode: queue up all children before visiting them.
  */
@@ -1473,6 +1491,15 @@ done:
 	return bftw_state_destroy(&state);
 }
 
+/** Select bftw_stream() or bftw_batch() appropriately. */
+static int bftw_auto(const struct bftw_args *args) {
+	if (args->flags & BFTW_SORT) {
+		return bftw_batch(args);
+	} else {
+		return bftw_stream(args);
+	}
+}
+
 /**
  * Iterative deepening search state.
  */
@@ -1483,10 +1510,14 @@ struct bftw_ids_state {
 	void *ptr;
 	/** Which visit this search corresponds to. */
 	enum bftw_visit visit;
-	/** The current target depth. */
-	size_t depth;
+	/** Whether to override the bftw_visit. */
+	bool force_visit;
+	/** The current minimum depth (inclusive). */
+	size_t min_depth;
+	/** The current maximum depth (exclusive). */
+	size_t max_depth;
 	/** The set of pruned paths. */
-	struct trie *pruned;
+	struct trie pruned;
 	/** An error code to report. */
 	int error;
 	/** Whether the bottom has been found. */
@@ -1499,40 +1530,46 @@ struct bftw_ids_state {
 static enum bftw_action bftw_ids_callback(const struct BFTW *ftwbuf, void *ptr) {
 	struct bftw_ids_state *state = ptr;
 
-	struct BFTW *mutbuf = (struct BFTW *)ftwbuf;
-	mutbuf->visit = state->visit;
+	if (state->force_visit) {
+		struct BFTW *mutbuf = (struct BFTW *)ftwbuf;
+		mutbuf->visit = state->visit;
+	}
 
-	if (ftwbuf->typeflag == BFTW_ERROR) {
-		if (state->depth - ftwbuf->depth <= 1) {
+	if (ftwbuf->type == BFTW_ERROR) {
+		if (ftwbuf->depth + 1 >= state->min_depth) {
 			return state->delegate(ftwbuf, state->ptr);
 		} else {
 			return BFTW_PRUNE;
 		}
 	}
 
-	if (ftwbuf->depth < state->depth) {
-		if (trie_find_str(state->pruned, ftwbuf->path)) {
+	if (ftwbuf->depth < state->min_depth) {
+		if (trie_find_str(&state->pruned, ftwbuf->path)) {
 			return BFTW_PRUNE;
 		} else {
 			return BFTW_CONTINUE;
 		}
 	} else if (state->visit == BFTW_POST) {
-		if (trie_find_str(state->pruned, ftwbuf->path)) {
+		if (trie_find_str(&state->pruned, ftwbuf->path)) {
 			return BFTW_PRUNE;
 		}
 	}
 
-	state->bottom = false;
-
-	enum bftw_action ret = state->delegate(ftwbuf, state->ptr);
+	enum bftw_action ret = BFTW_CONTINUE;
+	if (ftwbuf->visit == state->visit) {
+		ret = state->delegate(ftwbuf, state->ptr);
+	}
 
 	switch (ret) {
 	case BFTW_CONTINUE:
-		ret = BFTW_PRUNE;
+		if (ftwbuf->type == BFTW_DIR && ftwbuf->depth + 1 >= state->max_depth) {
+			state->bottom = false;
+			ret = BFTW_PRUNE;
+		}
 		break;
 	case BFTW_PRUNE:
-		if (ftwbuf->typeflag == BFTW_DIR) {
-			if (!trie_insert_str(state->pruned, ftwbuf->path)) {
+		if (ftwbuf->type == BFTW_DIR) {
+			if (!trie_insert_str(&state->pruned, ftwbuf->path)) {
 				state->error = errno;
 				state->quit = true;
 				ret = BFTW_STOP;
@@ -1547,66 +1584,123 @@ static enum bftw_action bftw_ids_callback(const struct BFTW *ftwbuf, void *ptr)
 	return ret;
 }
 
+/** Initialize iterative deepening state. */
+static void bftw_ids_init(const struct bftw_args *args, struct bftw_ids_state *state, struct bftw_args *ids_args) {
+	state->delegate = args->callback;
+	state->ptr = args->ptr;
+	state->visit = BFTW_PRE;
+	state->force_visit = false;
+	state->min_depth = 0;
+	state->max_depth = 1;
+	trie_init(&state->pruned);
+	state->error = 0;
+	state->bottom = false;
+	state->quit = false;
+
+	*ids_args = *args;
+	ids_args->callback = bftw_ids_callback;
+	ids_args->ptr = state;
+	ids_args->flags &= ~BFTW_POST_ORDER;
+	ids_args->strategy = BFTW_DFS;
+}
+
+/** Finish an iterative deepening search. */
+static int bftw_ids_finish(struct bftw_ids_state *state) {
+	int ret = 0;
+
+	if (state->error) {
+		ret = -1;
+	} else {
+		state->error = errno;
+	}
+
+	trie_destroy(&state->pruned);
+
+	errno = state->error;
+	return ret;
+}
+
 /**
  * Iterative deepening bftw() wrapper.
  */
 static int bftw_ids(const struct bftw_args *args) {
-	struct trie pruned;
-	trie_init(&pruned);
+	struct bftw_ids_state state;
+	struct bftw_args ids_args;
+	bftw_ids_init(args, &state, &ids_args);
 
-	struct bftw_ids_state state = {
-		.delegate = args->callback,
-		.ptr = args->ptr,
-		.visit = BFTW_PRE,
-		.depth = 0,
-		.pruned = &pruned,
-		.bottom = false,
-	};
-
-	struct bftw_args ids_args = *args;
-	ids_args.callback = bftw_ids_callback;
-	ids_args.ptr = &state;
-	ids_args.flags &= ~BFTW_DEPTH;
-	ids_args.strategy = BFTW_DFS;
-
-	int ret = 0;
-
-	while (ret == 0 && !state.quit && !state.bottom) {
+	while (!state.quit && !state.bottom) {
 		state.bottom = true;
-		ret = args->flags & BFTW_SORT ? bftw_batch(&ids_args) : bftw_stream(&ids_args);
-		++state.depth;
+
+		if (bftw_auto(&ids_args) != 0) {
+			state.error = errno;
+			state.quit = true;
+		}
+
+		++state.min_depth;
+		++state.max_depth;
 	}
 
-	if (args->flags & BFTW_DEPTH) {
+	if (args->flags & BFTW_POST_ORDER) {
 		state.visit = BFTW_POST;
+		state.force_visit = true;
 
-		while (ret == 0 && !state.quit && state.depth > 0) {
-			--state.depth;
-			ret = args->flags & BFTW_SORT ? bftw_batch(&ids_args) : bftw_stream(&ids_args);
+		while (!state.quit && state.min_depth > 0) {
+			--state.max_depth;
+			--state.min_depth;
+
+			if (bftw_auto(&ids_args) != 0) {
+				state.error = errno;
+				state.quit = true;
+			}
 		}
 	}
 
-	if (state.error) {
-		ret = -1;
-	} else {
-		state.error = errno;
+	return bftw_ids_finish(&state);
+}
+
+/**
+ * Exponential deepening bftw() wrapper.
+ */
+static int bftw_eds(const struct bftw_args *args) {
+	struct bftw_ids_state state;
+	struct bftw_args ids_args;
+	bftw_ids_init(args, &state, &ids_args);
+
+	while (!state.quit && !state.bottom) {
+		state.bottom = true;
+
+		if (bftw_auto(&ids_args) != 0) {
+			state.error = errno;
+			state.quit = true;
+		}
+
+		state.min_depth = state.max_depth;
+		state.max_depth *= 2;
 	}
-	trie_destroy(&pruned);
-	errno = state.error;
-	return ret;
+
+	if (!state.quit && (args->flags & BFTW_POST_ORDER)) {
+		state.visit = BFTW_POST;
+		state.min_depth = 0;
+		ids_args.flags |= BFTW_POST_ORDER;
+
+		if (bftw_auto(&ids_args) != 0) {
+			state.error = errno;
+		}
+	}
+
+	return bftw_ids_finish(&state);
 }
 
 int bftw(const struct bftw_args *args) {
 	switch (args->strategy) {
 	case BFTW_BFS:
-		if (!(args->flags & BFTW_SORT)) {
-			return bftw_stream(args);
-		}
-		// Fallthrough
+		return bftw_auto(args);
 	case BFTW_DFS:
 		return bftw_batch(args);
 	case BFTW_IDS:
 		return bftw_ids(args);
+	case BFTW_EDS:
+		return bftw_eds(args);
 	}
 
 	errno = EINVAL;
