@@ -17,6 +17,7 @@
 #include "color.h"
 #include "bftw.h"
 #include "dstring.h"
+#include "expr.h"
 #include "fsade.h"
 #include "stat.h"
 #include "trie.h"
@@ -555,16 +556,16 @@ static bool is_link_broken(const struct BFTW *ftwbuf) {
 }
 
 /** Get the color for a file. */
-static const char *file_color(const struct colors *colors, const char *filename, const struct BFTW *ftwbuf, enum bfs_stat_flag flags) {
-	enum bftw_typeflag typeflag = bftw_typeflag(ftwbuf, flags);
-	if (typeflag == BFTW_ERROR) {
+static const char *file_color(const struct colors *colors, const char *filename, const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
+	enum bftw_type type = bftw_type(ftwbuf, flags);
+	if (type == BFTW_ERROR) {
 		goto error;
 	}
 
 	const struct bfs_stat *statbuf = NULL;
 	const char *color = NULL;
 
-	switch (typeflag) {
+	switch (type) {
 	case BFTW_REG:
 		if (colors->setuid || colors->setgid || colors->executable || colors->multi_hard) {
 			statbuf = bftw_stat(ftwbuf, flags);
@@ -715,10 +716,10 @@ static int print_colored(const struct colors *colors, const char *esc, const cha
 }
 
 /** Find the offset of the first broken path component. */
-static ssize_t first_broken_offset(const char *path, const struct BFTW *ftwbuf, enum bfs_stat_flag flags, size_t max) {
+static ssize_t first_broken_offset(const char *path, const struct BFTW *ftwbuf, enum bfs_stat_flags flags, size_t max) {
 	ssize_t ret = max;
 
-	if (bftw_typeflag(ftwbuf, flags) != BFTW_ERROR) {
+	if (bftw_type(ftwbuf, flags) != BFTW_ERROR) {
 		goto out;
 	}
 
@@ -774,7 +775,7 @@ out:
 }
 
 /** Print the directories leading up to a file. */
-static int print_dirs_colored(CFILE *cfile, const char *path, const struct BFTW *ftwbuf, enum bfs_stat_flag flags, size_t nameoff) {
+static int print_dirs_colored(CFILE *cfile, const char *path, const struct BFTW *ftwbuf, enum bfs_stat_flags flags, size_t nameoff) {
 	const struct colors *colors = cfile->colors;
 	FILE *file = cfile->file;
 
@@ -789,7 +790,7 @@ static int print_dirs_colored(CFILE *cfile, const char *path, const struct BFTW
 		}
 	}
 
-	if (broken < nameoff) {
+	if ((size_t)broken < nameoff) {
 		const char *color = colors->missing;
 		if (!color) {
 			color = colors->orphan;
@@ -803,7 +804,7 @@ static int print_dirs_colored(CFILE *cfile, const char *path, const struct BFTW
 }
 
 /** Print a path with colors. */
-static int print_path_colored(CFILE *cfile, const char *path, const struct BFTW *ftwbuf, enum bfs_stat_flag flags) {
+static int print_path_colored(CFILE *cfile, const char *path, const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
 	const struct colors *colors = cfile->colors;
 	FILE *file = cfile->file;
 
@@ -828,8 +829,8 @@ static int print_path(CFILE *cfile, const struct BFTW *ftwbuf) {
 		return fputs(ftwbuf->path, cfile->file) == EOF ? -1 : 0;
 	}
 
-	enum bfs_stat_flag flags = ftwbuf->stat_flags;
-	if (colors && colors->link_as_target && ftwbuf->typeflag == BFTW_LNK) {
+	enum bfs_stat_flags flags = ftwbuf->stat_flags;
+	if (colors && colors->link_as_target && ftwbuf->type == BFTW_LNK) {
 		flags = BFS_STAT_TRYFOLLOW;
 	}
 
@@ -840,11 +841,8 @@ static int print_path(CFILE *cfile, const struct BFTW *ftwbuf) {
 static int print_link_target(CFILE *cfile, const struct BFTW *ftwbuf) {
 	int ret = -1;
 
-	size_t len = 0;
-	const struct bfs_stat *statbuf = bftw_stat(ftwbuf, BFS_STAT_NOFOLLOW);
-	if (statbuf) {
-		len = statbuf->size;
-	}
+	const struct bfs_stat *statbuf = bftw_cached_stat(ftwbuf, BFS_STAT_NOFOLLOW);
+	size_t len = statbuf ? statbuf->size : 0;
 
 	char *target = xreadlinkat(ftwbuf->at_fd, ftwbuf->at_path, len);
 	if (!target) {
@@ -863,6 +861,65 @@ done:
 	return ret;
 }
 
+/** Dump a parsed expression tree, for debugging. */
+static int print_expr(CFILE *cfile, const struct expr *expr, bool verbose) {
+	if (fputs("(", cfile->file) == EOF) {
+		return -1;
+	}
+
+	if (expr->lhs || expr->rhs) {
+		if (cfprintf(cfile, "${red}%s${rs}", expr->argv[0]) < 0) {
+			return -1;
+		}
+	} else {
+		if (cfprintf(cfile, "${blu}%s${rs}", expr->argv[0]) < 0) {
+			return -1;
+		}
+	}
+
+	for (size_t i = 1; i < expr->argc; ++i) {
+		if (cfprintf(cfile, " ${bld}%s${rs}", expr->argv[i]) < 0) {
+			return -1;
+		}
+	}
+
+	if (verbose) {
+		double rate = 0.0, time = 0.0;
+		if (expr->evaluations) {
+			rate = 100.0*expr->successes/expr->evaluations;
+			time = (1.0e9*expr->elapsed.tv_sec + expr->elapsed.tv_nsec)/expr->evaluations;
+		}
+		if (cfprintf(cfile, " [${ylw}%zu${rs}/${ylw}%zu${rs}=${ylw}%g%%${rs}; ${ylw}%gns${rs}]",
+		             expr->successes, expr->evaluations, rate, time)) {
+			return -1;
+		}
+	}
+
+	if (expr->lhs) {
+		if (fputs(" ", cfile->file) == EOF) {
+			return -1;
+		}
+		if (print_expr(cfile, expr->lhs, verbose) != 0) {
+			return -1;
+		}
+	}
+
+	if (expr->rhs) {
+		if (fputs(" ", cfile->file) == EOF) {
+			return -1;
+		}
+		if (print_expr(cfile, expr->rhs, verbose) != 0) {
+			return -1;
+		}
+	}
+
+	if (fputs(")", cfile->file) == EOF) {
+		return -1;
+	}
+
+	return 0;
+}
+
 int cfprintf(CFILE *cfile, const char *format, ...) {
 	va_list args;
 	va_start(args, format);
@@ -946,6 +1003,17 @@ int cvfprintf(CFILE *cfile, const char *format, va_list args) {
 					}
 					break;
 
+				case 'e':
+					if (print_expr(cfile, va_arg(args, const struct expr *), false) != 0) {
+						return -1;
+					}
+					break;
+				case 'E':
+					if (print_expr(cfile, va_arg(args, const struct expr *), true) != 0) {
+						return -1;
+					}
+					break;
+
 				default:
 					goto invalid;
 				}
