@@ -37,7 +37,7 @@ static struct dstring *dstrheader(const char *dstr) {
 
 /** Get the correct size for a dstring with the given capacity. */
 static size_t dstrsize(size_t capacity) {
-	return sizeof(struct dstring) + capacity + 1;
+	return BFS_FLEX_SIZEOF(struct dstring, data, capacity + 1);
 }
 
 /** Allocate a dstring with the given contents. */
@@ -132,25 +132,47 @@ char *dstrprintf(const char *format, ...) {
 	va_list args;
 
 	va_start(args, format);
-	int len = vsnprintf(NULL, 0, format, args);
+	char *str = dstrvprintf(format, args);
 	va_end(args);
 
-	assert(len > 0);
+	return str;
+}
 
-	char *str = dstralloc(len);
+char *dstrvprintf(const char *format, va_list args) {
+	// Guess a length to try to avoid calling vsnprintf() twice
+	int len, cap = 2*strlen(format);
+	char *str = dstralloc(cap);
 	if (!str) {
 		return NULL;
 	}
 
-	va_start(args, format);
-	len = vsnprintf(str, len + 1, format, args);
-	va_end(args);
+	va_list args2;
+	va_copy(args2, args);
+
+	len = vsnprintf(str, cap + 1, format, args);
+	if (len > cap) {
+		if (dstreserve(&str, len) != 0) {
+			goto fail;
+		}
+
+		cap = len;
+		len = vsnprintf(str, cap + 1, format, args2);
+		assert(len == cap);
+	}
+
+	va_end(args2);
+
+	if (len < 0) {
+		goto fail;
+	}
 
 	struct dstring *header = dstrheader(str);
-	assert(len == header->capacity);
 	header->length = len;
-
 	return str;
+
+fail:
+	dstrfree(str);
+	return NULL;
 }
 
 void dstrfree(char *dstr) {
