@@ -19,13 +19,14 @@
  */
 
 #include "eval.h"
+#include "bar.h"
 #include "bftw.h"
-#include "cmdline.h"
 #include "color.h"
 #include "darray.h"
 #include "diag.h"
 #include "dstring.h"
 #include "exec.h"
+#include "expr.h"
 #include "fsade.h"
 #include "mtab.h"
 #include "printf.h"
@@ -50,15 +51,16 @@
 #include <sys/stat.h>
 #include <time.h>
 #include <unistd.h>
+#include <wchar.h>
 
 struct eval_state {
 	/** Data about the current file. */
 	const struct BFTW *ftwbuf;
-	/** The parsed command line. */
-	const struct cmdline *cmdline;
+	/** The bfs context. */
+	const struct bfs_ctx *ctx;
 	/** The bftw() callback return value. */
 	enum bftw_action action;
-	/** The eval_cmdline() return value. */
+	/** The bfs_eval() return value. */
 	int *ret;
 	/** Whether to quit immediately. */
 	bool quit;
@@ -70,10 +72,10 @@ struct eval_state {
 BFS_FORMATTER(2, 3)
 static void eval_error(struct eval_state *state, const char *format, ...) {
 	int error = errno;
-	const struct cmdline *cmdline = state->cmdline;
-	CFILE *cerr = cmdline->cerr;
+	const struct bfs_ctx *ctx = state->ctx;
+	CFILE *cerr = ctx->cerr;
 
-	bfs_error(cmdline, "%pP: ", state->ftwbuf);
+	bfs_error(ctx, "%pP: ", state->ftwbuf);
 
 	va_list args;
 	va_start(args, format);
@@ -86,7 +88,7 @@ static void eval_error(struct eval_state *state, const char *format, ...) {
  * Check if an error should be ignored.
  */
 static bool eval_should_ignore(const struct eval_state *state, int error) {
-	return state->cmdline->ignore_races
+	return state->ctx->ignore_races
 		&& is_nonexistence_error(error)
 		&& state->ftwbuf->depth > 0;
 }
@@ -295,7 +297,13 @@ bool eval_nogroup(const struct expr *expr, struct eval_state *state) {
 		return false;
 	}
 
-	return bfs_getgrgid(state->cmdline->groups, statbuf->gid) == NULL;
+	const struct bfs_groups *groups = bfs_ctx_groups(state->ctx);
+	if (!groups) {
+		eval_report_error(state);
+		return false;
+	}
+
+	return bfs_getgrgid(groups, statbuf->gid) == NULL;
 }
 
 /**
@@ -307,7 +315,13 @@ bool eval_nouser(const struct expr *expr, struct eval_state *state) {
 		return false;
 	}
 
-	return bfs_getpwuid(state->cmdline->users, statbuf->uid) == NULL;
+	const struct bfs_users *users = bfs_ctx_users(state->ctx);
+	if (!users) {
+		eval_report_error(state);
+		return false;
+	}
+
+	return bfs_getpwuid(users, statbuf->uid) == NULL;
 }
 
 /**
@@ -324,7 +338,7 @@ bool eval_delete(const struct expr *expr, struct eval_state *state) {
 	int flag = 0;
 
 	// We need to know the actual type of the path, not what it points to
-	enum bftw_typeflag type = bftw_typeflag(ftwbuf, BFS_STAT_NOFOLLOW);
+	enum bftw_type type = bftw_type(ftwbuf, BFS_STAT_NOFOLLOW);
 	if (type == BFTW_DIR) {
 		flag |= AT_REMOVEDIR;
 	} else if (type == BFTW_ERROR) {
@@ -341,18 +355,18 @@ bool eval_delete(const struct expr *expr, struct eval_state *state) {
 }
 
 /** Finish any pending -exec ... + operations. */
-static int eval_exec_finish(const struct expr *expr, const struct cmdline *cmdline) {
+static int eval_exec_finish(const struct expr *expr, const struct bfs_ctx *ctx) {
 	int ret = 0;
 	if (expr->execbuf && bfs_exec_finish(expr->execbuf) != 0) {
 		if (errno != 0) {
-			bfs_error(cmdline, "%s %s: %m.\n", expr->argv[0], expr->argv[1]);
+			bfs_error(ctx, "%s %s: %m.\n", expr->argv[0], expr->argv[1]);
 		}
 		ret = -1;
 	}
-	if (expr->lhs && eval_exec_finish(expr->lhs, cmdline) != 0) {
+	if (expr->lhs && eval_exec_finish(expr->lhs, ctx) != 0) {
 		ret = -1;
 	}
-	if (expr->rhs && eval_exec_finish(expr->rhs, cmdline) != 0) {
+	if (expr->rhs && eval_exec_finish(expr->rhs, ctx) != 0) {
 		ret = -1;
 	}
 	return ret;
@@ -394,7 +408,7 @@ bool eval_empty(const struct expr *expr, struct eval_state *state) {
 	bool ret = false;
 	const struct BFTW *ftwbuf = state->ftwbuf;
 
-	if (ftwbuf->typeflag == BFTW_DIR) {
+	if (ftwbuf->type == BFTW_DIR) {
 		int dfd = openat(ftwbuf->at_fd, ftwbuf->at_path, O_RDONLY | O_CLOEXEC | O_DIRECTORY);
 		if (dfd < 0) {
 			eval_report_error(state);
@@ -416,7 +430,7 @@ bool eval_empty(const struct expr *expr, struct eval_state *state) {
 		}
 
 		closedir(dir);
-	} else if (ftwbuf->typeflag == BFTW_REG) {
+	} else if (ftwbuf->type == BFTW_REG) {
 		const struct bfs_stat *statbuf = eval_stat(state);
 		if (statbuf) {
 			ret = statbuf->size == 0;
@@ -436,7 +450,13 @@ bool eval_fstype(const struct expr *expr, struct eval_state *state) {
 		return false;
 	}
 
-	const char *type = bfs_fstype(state->cmdline->mtab, statbuf);
+	const struct bfs_mtab *mtab = bfs_ctx_mtab(state->ctx);
+	if (!mtab) {
+		eval_report_error(state);
+		return false;
+	}
+
+	const char *type = bfs_fstype(mtab, statbuf);
 	return strcmp(type, expr->sdata) == 0;
 }
 
@@ -448,18 +468,6 @@ bool eval_hidden(const struct expr *expr, struct eval_state *state) {
 	return ftwbuf->nameoff > 0 && ftwbuf->path[ftwbuf->nameoff] == '.';
 }
 
-/**
- * -nohidden action.
- */
-bool eval_nohidden(const struct expr *expr, struct eval_state *state) {
-	if (eval_hidden(expr, state)) {
-		eval_prune(expr, state);
-		return false;
-	} else {
-		return true;
-	}
-}
-
 /**
  * -inum test.
  */
@@ -492,16 +500,14 @@ bool eval_lname(const struct expr *expr, struct eval_state *state) {
 	char *name = NULL;
 
 	const struct BFTW *ftwbuf = state->ftwbuf;
-	if (ftwbuf->typeflag != BFTW_LNK) {
+	if (ftwbuf->type != BFTW_LNK) {
 		goto done;
 	}
 
-	const struct bfs_stat *statbuf = eval_stat(state);
-	if (!statbuf) {
-		goto done;
-	}
+	const struct bfs_stat *statbuf = bftw_cached_stat(ftwbuf, BFS_STAT_NOFOLLOW);
+	size_t len = statbuf ? statbuf->size : 0;
 
-	name = xreadlinkat(ftwbuf->at_fd, ftwbuf->at_path, statbuf->size);
+	name = xreadlinkat(ftwbuf->at_fd, ftwbuf->at_path, len);
 	if (!name) {
 		eval_report_error(state);
 		goto done;
@@ -560,7 +566,7 @@ bool eval_perm(const struct expr *expr, struct eval_state *state) {
 
 	mode_t mode = statbuf->mode;
 	mode_t target;
-	if (state->ftwbuf->typeflag == BFTW_DIR) {
+	if (state->ftwbuf->type == BFTW_DIR) {
 		target = expr->dir_mode;
 	} else {
 		target = expr->file_mode;
@@ -586,8 +592,8 @@ bool eval_perm(const struct expr *expr, struct eval_state *state) {
 bool eval_fls(const struct expr *expr, struct eval_state *state) {
 	CFILE *cfile = expr->cfile;
 	FILE *file = cfile->file;
-	const struct bfs_users *users = state->cmdline->users;
-	const struct bfs_groups *groups = state->cmdline->groups;
+	const struct bfs_users *users = bfs_ctx_users(state->ctx);
+	const struct bfs_groups *groups = bfs_ctx_groups(state->ctx);
 	const struct BFTW *ftwbuf = state->ftwbuf;
 	const struct bfs_stat *statbuf = eval_stat(state);
 	if (!statbuf) {
@@ -597,7 +603,7 @@ bool eval_fls(const struct expr *expr, struct eval_state *state) {
 	uintmax_t ino = statbuf->ino;
 	uintmax_t blocks = ((uintmax_t)statbuf->blocks*BFS_STAT_BLKSIZE + 1023)/1024;
 	char mode[11];
-	format_mode(statbuf->mode, mode);
+	xstrmode(statbuf->mode, mode);
 	char acl = bfs_check_acl(ftwbuf) > 0 ? '+' : ' ';
 	uintmax_t nlink = statbuf->nlink;
 	if (fprintf(file, "%9ju %6ju %s%c %2ju ", ino, blocks, mode, acl, nlink) < 0) {
@@ -628,7 +634,7 @@ bool eval_fls(const struct expr *expr, struct eval_state *state) {
 		}
 	}
 
-	if (ftwbuf->typeflag & (BFTW_BLK | BFTW_CHR)) {
+	if (ftwbuf->type == BFTW_BLK || ftwbuf->type == BFTW_CHR) {
 		int ma = bfs_major(statbuf->rdev);
 		int mi = bfs_minor(statbuf->rdev);
 		if (fprintf(file, " %3d, %3d", ma, mi) < 0) {
@@ -666,7 +672,7 @@ bool eval_fls(const struct expr *expr, struct eval_state *state) {
 		goto error;
 	}
 
-	if (ftwbuf->typeflag == BFTW_LNK) {
+	if (ftwbuf->type == BFTW_LNK) {
 		if (cfprintf(cfile, " -> %pL", ftwbuf) < 0) {
 			goto error;
 		}
@@ -789,14 +795,14 @@ bool eval_regex(const struct expr *expr, struct eval_state *state) {
 #endif
 	int err = regexec(expr->regex, path, 1, &match, flags);
 	if (err == 0) {
-		return match.rm_so == 0 && match.rm_eo == len;
+		return match.rm_so == 0 && (size_t)match.rm_eo == len;
 	} else if (err != REG_NOMATCH) {
 		char *str = xregerror(err, expr->regex);
 		if (str) {
 			eval_error(state, "%s.\n", str);
 			free(str);
 		} else {
-			perror("xregerror()");
+			eval_error(state, "xregerror(): %m.\n");
 		}
 
 		*state->ret = EXIT_FAILURE;
@@ -859,7 +865,7 @@ bool eval_sparse(const struct expr *expr, struct eval_state *state) {
  * -type test.
  */
 bool eval_type(const struct expr *expr, struct eval_state *state) {
-	return state->ftwbuf->typeflag & expr->idata;
+	return (1 << state->ftwbuf->type) & expr->idata;
 }
 
 /**
@@ -875,18 +881,31 @@ bool eval_xattr(const struct expr *expr, struct eval_state *state) {
 	}
 }
 
+/**
+ * -xattr test.
+ */
+bool eval_xattrname(const struct expr *expr, struct eval_state *state) {
+	int ret = bfs_check_xattr_named(state->ftwbuf, expr->sdata);
+	if (ret >= 0) {
+		return ret;
+	} else {
+		eval_report_error(state);
+		return false;
+	}
+}
+
 /**
  * -xtype test.
  */
 bool eval_xtype(const struct expr *expr, struct eval_state *state) {
 	const struct BFTW *ftwbuf = state->ftwbuf;
-	enum bfs_stat_flag flags = ftwbuf->stat_flags ^ (BFS_STAT_NOFOLLOW | BFS_STAT_TRYFOLLOW);
-	enum bftw_typeflag type = bftw_typeflag(ftwbuf, flags);
+	enum bfs_stat_flags flags = ftwbuf->stat_flags ^ (BFS_STAT_NOFOLLOW | BFS_STAT_TRYFOLLOW);
+	enum bftw_type type = bftw_type(ftwbuf, flags);
 	if (type == BFTW_ERROR) {
 		eval_report_error(state);
 		return false;
 	} else {
-		return type & expr->idata;
+		return (1 << type) & expr->idata;
 	}
 }
 
@@ -899,11 +918,11 @@ bool eval_xtype(const struct expr *expr, struct eval_state *state) {
 /**
  * Call clock_gettime(), if available.
  */
-static int eval_gettime(struct timespec *ts) {
+static int eval_gettime(struct eval_state *state, struct timespec *ts) {
 #ifdef BFS_CLOCK
 	int ret = clock_gettime(BFS_CLOCK, ts);
 	if (ret != 0) {
-		perror("clock_gettime()");
+		bfs_warning(state->ctx, "%pP: clock_gettime(): %m.\n", state->ftwbuf);
 	}
 	return ret;
 #else
@@ -912,17 +931,17 @@ static int eval_gettime(struct timespec *ts) {
 }
 
 /**
- * Record the time that elapsed evaluating an expression.
+ * Record an elapsed time.
  */
-static void add_elapsed(struct expr *expr, const struct timespec *start, const struct timespec *end) {
-	expr->elapsed.tv_sec += end->tv_sec - start->tv_sec;
-	expr->elapsed.tv_nsec += end->tv_nsec - start->tv_nsec;
-	if (expr->elapsed.tv_nsec < 0) {
-		expr->elapsed.tv_nsec += 1000000000L;
-		--expr->elapsed.tv_sec;
-	} else if (expr->elapsed.tv_nsec >= 1000000000L) {
-		expr->elapsed.tv_nsec -= 1000000000L;
-		++expr->elapsed.tv_sec;
+static void timespec_elapsed(struct timespec *elapsed, const struct timespec *start, const struct timespec *end) {
+	elapsed->tv_sec += end->tv_sec - start->tv_sec;
+	elapsed->tv_nsec += end->tv_nsec - start->tv_nsec;
+	if (elapsed->tv_nsec < 0) {
+		elapsed->tv_nsec += 1000000000L;
+		--elapsed->tv_sec;
+	} else if (elapsed->tv_nsec >= 1000000000L) {
+		elapsed->tv_nsec -= 1000000000L;
+		++elapsed->tv_sec;
 	}
 }
 
@@ -931,9 +950,9 @@ static void add_elapsed(struct expr *expr, const struct timespec *start, const s
  */
 static bool eval_expr(struct expr *expr, struct eval_state *state) {
 	struct timespec start, end;
-	bool time = state->cmdline->debug & DEBUG_RATES;
+	bool time = state->ctx->debug & DEBUG_RATES;
 	if (time) {
-		if (eval_gettime(&start) != 0) {
+		if (eval_gettime(state, &start) != 0) {
 			time = false;
 		}
 	}
@@ -943,8 +962,8 @@ static bool eval_expr(struct expr *expr, struct eval_state *state) {
 	bool ret = expr->eval(expr, state);
 
 	if (time) {
-		if (eval_gettime(&end) == 0) {
-			add_elapsed(expr, &start, &end);
+		if (eval_gettime(state, &end) == 0) {
+			timespec_elapsed(&expr->elapsed, &start, &end);
 		}
 	}
 
@@ -1013,6 +1032,110 @@ bool eval_comma(const struct expr *expr, struct eval_state *state) {
 	return eval_expr(expr->rhs, state);
 }
 
+/** Update the status bar. */
+static void eval_status(struct eval_state *state, struct bfs_bar *bar, struct timespec *last_status, size_t count) {
+	struct timespec now;
+	if (eval_gettime(state, &now) == 0) {
+		struct timespec elapsed = {0};
+		timespec_elapsed(&elapsed, last_status, &now);
+
+		// Update every 0.1s
+		if (elapsed.tv_sec > 0 || elapsed.tv_nsec >= 100000000L) {
+			*last_status = now;
+		} else {
+			return;
+		}
+	}
+
+	size_t width = bfs_bar_width(bar);
+	if (width < 3) {
+		return;
+	}
+
+	const struct BFTW *ftwbuf = state->ftwbuf;
+
+	char *rhs = dstrprintf(" (visited: %zu, depth: %2zu)", count, ftwbuf->depth);
+	if (!rhs) {
+		return;
+	}
+
+	size_t rhslen = dstrlen(rhs);
+	if (3 + rhslen > width) {
+		dstresize(&rhs, 0);
+		rhslen = 0;
+	}
+
+	char *status = dstralloc(0);
+	if (!status) {
+		goto out_rhs;
+	}
+
+	const char *path = ftwbuf->path;
+	size_t pathlen = ftwbuf->nameoff;
+	if (ftwbuf->depth == 0) {
+		pathlen = strlen(path);
+	}
+
+	// Try to make sure even wide characters fit in the status bar
+	size_t pathmax = width - rhslen - 3;
+	size_t pathwidth = 0;
+	mbstate_t mb;
+	memset(&mb, 0, sizeof(mb));
+	while (pathlen > 0) {
+		wchar_t wc;
+		size_t len = mbrtowc(&wc, path, pathlen, &mb);
+		int width;
+		if (len == (size_t)-1) {
+			// Invalid byte sequence, assume a single-width ?
+			len = 1;
+			width = 1;
+			memset(&mb, 0, sizeof(mb));
+		} else if (len == (size_t)-2) {
+			// Incomplete byte sequence, assume a single-width ?
+			len = pathlen;
+			width = 1;
+		} else {
+			width = wcwidth(wc);
+			if (width < 0) {
+				width = 0;
+			}
+		}
+
+		if (pathwidth + width > pathmax) {
+			break;
+		}
+
+		if (dstrncat(&status, path, len) != 0) {
+			goto out_rhs;
+		}
+
+		path += len;
+		pathlen -= len;
+		pathwidth += width;
+	}
+
+	if (dstrcat(&status, "...") != 0) {
+		goto out_rhs;
+	}
+
+	while (pathwidth < pathmax) {
+		if (dstrapp(&status, ' ') != 0) {
+			goto out_rhs;
+		}
+		++pathwidth;
+	}
+
+	if (dstrcat(&status, rhs) != 0) {
+		goto out_rhs;
+	}
+
+	bfs_bar_update(bar, status);
+
+	dstrfree(status);
+out_rhs:
+	dstrfree(rhs);
+}
+
 /** Check if we've seen a file before. */
 static bool eval_file_unique(struct eval_state *state, struct trie *seen) {
 	const struct bfs_stat *statbuf = eval_stat(state);
@@ -1052,7 +1175,9 @@ static bool eval_file_unique(struct eval_state *state, struct trie *seen) {
 /**
  * Log a stat() call.
  */
-static void debug_stat(const struct BFTW *ftwbuf, const struct bftw_stat *cache, enum bfs_stat_flag flags) {
+static void debug_stat(const struct bfs_ctx *ctx, const struct BFTW *ftwbuf, const struct bftw_stat *cache, enum bfs_stat_flags flags) {
+	bfs_debug_prefix(ctx, DEBUG_STAT);
+
 	fprintf(stderr, "bfs_stat(");
 	if (ftwbuf->at_fd == AT_FDCWD) {
 		fprintf(stderr, "AT_FDCWD");
@@ -1081,47 +1206,49 @@ static void debug_stat(const struct BFTW *ftwbuf, const struct bftw_stat *cache,
 /**
  * Log any stat() calls that happened.
  */
-static void debug_stats(const struct BFTW *ftwbuf) {
+static void debug_stats(const struct bfs_ctx *ctx, const struct BFTW *ftwbuf) {
+	if (!(ctx->debug & DEBUG_STAT)) {
+		return;
+	}
+
 	const struct bfs_stat *statbuf = ftwbuf->stat_cache.buf;
 	if (statbuf || ftwbuf->stat_cache.error) {
-		debug_stat(ftwbuf, &ftwbuf->stat_cache, BFS_STAT_FOLLOW);
+		debug_stat(ctx, ftwbuf, &ftwbuf->stat_cache, BFS_STAT_FOLLOW);
 	}
 
 	const struct bfs_stat *lstatbuf = ftwbuf->lstat_cache.buf;
 	if ((lstatbuf && lstatbuf != statbuf) || ftwbuf->lstat_cache.error) {
-		debug_stat(ftwbuf, &ftwbuf->lstat_cache, BFS_STAT_NOFOLLOW);
-	}
-}
-
-/**
- * Dump the bftw_typeflag for -D search.
- */
-static const char *dump_bftw_typeflag(enum bftw_typeflag type) {
-#define DUMP_BFTW_TYPEFLAG_CASE(flag)		\
-	case flag:				\
-		return #flag
-
-	switch (type) {
-		DUMP_BFTW_TYPEFLAG_CASE(BFTW_BLK);
-		DUMP_BFTW_TYPEFLAG_CASE(BFTW_CHR);
-		DUMP_BFTW_TYPEFLAG_CASE(BFTW_DIR);
-		DUMP_BFTW_TYPEFLAG_CASE(BFTW_DOOR);
-		DUMP_BFTW_TYPEFLAG_CASE(BFTW_FIFO);
-		DUMP_BFTW_TYPEFLAG_CASE(BFTW_LNK);
-		DUMP_BFTW_TYPEFLAG_CASE(BFTW_PORT);
-		DUMP_BFTW_TYPEFLAG_CASE(BFTW_REG);
-		DUMP_BFTW_TYPEFLAG_CASE(BFTW_SOCK);
-		DUMP_BFTW_TYPEFLAG_CASE(BFTW_WHT);
-
-		DUMP_BFTW_TYPEFLAG_CASE(BFTW_ERROR);
-
-	default:
-		DUMP_BFTW_TYPEFLAG_CASE(BFTW_UNKNOWN);
+		debug_stat(ctx, ftwbuf, &ftwbuf->lstat_cache, BFS_STAT_NOFOLLOW);
 	}
 }
 
 #define DUMP_BFTW_MAP(value) [value] = #value
 
+/**
+ * Dump the bftw_type for -D search.
+ */
+static const char *dump_bftw_type(enum bftw_type type) {
+	static const char *types[] = {
+		DUMP_BFTW_MAP(BFTW_UNKNOWN),
+		DUMP_BFTW_MAP(BFTW_BLK),
+		DUMP_BFTW_MAP(BFTW_CHR),
+		DUMP_BFTW_MAP(BFTW_DIR),
+		DUMP_BFTW_MAP(BFTW_DOOR),
+		DUMP_BFTW_MAP(BFTW_FIFO),
+		DUMP_BFTW_MAP(BFTW_LNK),
+		DUMP_BFTW_MAP(BFTW_PORT),
+		DUMP_BFTW_MAP(BFTW_REG),
+		DUMP_BFTW_MAP(BFTW_SOCK),
+		DUMP_BFTW_MAP(BFTW_WHT),
+	};
+
+	if (type == BFTW_ERROR) {
+		return "BFTW_ERROR";
+	} else {
+		return types[type];
+	}
+}
+
 /**
  * Dump the bftw_visit for -D search.
  */
@@ -1149,30 +1276,44 @@ static const char *dump_bftw_action(enum bftw_action action) {
  * Type passed as the argument to the bftw() callback.
  */
 struct callback_args {
-	/** The parsed command line. */
-	const struct cmdline *cmdline;
+	/** The bfs context. */
+	const struct bfs_ctx *ctx;
+
+	/** The status bar. */
+	struct bfs_bar *bar;
+	/** The time of the last status update. */
+	struct timespec last_status;
+	/** The number of files visited so far. */
+	size_t count;
+
 	/** The set of seen files. */
 	struct trie *seen;
-	/** Eventual return value from eval_cmdline(). */
+
+	/** Eventual return value from bfs_eval(). */
 	int ret;
 };
 
 /**
  * bftw() callback.
  */
-static enum bftw_action cmdline_callback(const struct BFTW *ftwbuf, void *ptr) {
+static enum bftw_action eval_callback(const struct BFTW *ftwbuf, void *ptr) {
 	struct callback_args *args = ptr;
+	++args->count;
 
-	const struct cmdline *cmdline = args->cmdline;
+	const struct bfs_ctx *ctx = args->ctx;
 
 	struct eval_state state;
 	state.ftwbuf = ftwbuf;
-	state.cmdline = cmdline;
+	state.ctx = ctx;
 	state.action = BFTW_CONTINUE;
 	state.ret = &args->ret;
 	state.quit = false;
 
-	if (ftwbuf->typeflag == BFTW_ERROR) {
+	if (args->bar) {
+		eval_status(&state, args->bar, &args->last_status, args->count);
+	}
+
+	if (ftwbuf->type == BFTW_ERROR) {
 		if (!eval_should_ignore(&state, ftwbuf->error)) {
 			args->ret = EXIT_FAILURE;
 			eval_error(&state, "%s.\n", strerror(ftwbuf->error));
@@ -1181,49 +1322,51 @@ static enum bftw_action cmdline_callback(const struct BFTW *ftwbuf, void *ptr) {
 		goto done;
 	}
 
-	if (cmdline->unique && ftwbuf->visit == BFTW_PRE) {
+	if (ctx->unique && ftwbuf->visit == BFTW_PRE) {
 		if (!eval_file_unique(&state, args->seen)) {
 			goto done;
 		}
 	}
 
-	if (cmdline->xargs_safe && strpbrk(ftwbuf->path, " \t\n\'\"\\")) {
+	if (eval_expr(ctx->exclude, &state)) {
+		state.action = BFTW_PRUNE;
+		goto done;
+	}
+
+	if (ctx->xargs_safe && strpbrk(ftwbuf->path, " \t\n\'\"\\")) {
 		args->ret = EXIT_FAILURE;
 		eval_error(&state, "Path is not safe for xargs.\n");
 		state.action = BFTW_PRUNE;
 		goto done;
 	}
 
-	if (cmdline->maxdepth < 0 || ftwbuf->depth >= cmdline->maxdepth) {
+	if (ctx->maxdepth < 0 || ftwbuf->depth >= (size_t)ctx->maxdepth) {
 		state.action = BFTW_PRUNE;
 	}
 
 	// In -depth mode, only handle directories on the BFTW_POST visit
 	enum bftw_visit expected_visit = BFTW_PRE;
-	if ((cmdline->flags & BFTW_DEPTH)
-	    && (cmdline->strategy == BFTW_IDS || ftwbuf->typeflag == BFTW_DIR)
-	    && ftwbuf->depth < cmdline->maxdepth) {
+	if ((ctx->flags & BFTW_POST_ORDER)
+	    && (ctx->strategy == BFTW_IDS || ftwbuf->type == BFTW_DIR)
+	    && ftwbuf->depth < (size_t)ctx->maxdepth) {
 		expected_visit = BFTW_POST;
 	}
 
 	if (ftwbuf->visit == expected_visit
-	    && ftwbuf->depth >= cmdline->mindepth
-	    && ftwbuf->depth <= cmdline->maxdepth) {
-		eval_expr(cmdline->expr, &state);
+	    && ftwbuf->depth >= (size_t)ctx->mindepth
+	    && ftwbuf->depth <= (size_t)ctx->maxdepth) {
+		eval_expr(ctx->expr, &state);
 	}
 
 done:
-	if (cmdline->debug & DEBUG_STAT) {
-		debug_stats(ftwbuf);
-	}
+	debug_stats(ctx, ftwbuf);
 
-	if (cmdline->debug & DEBUG_SEARCH) {
-		fprintf(stderr, "cmdline_callback({\n");
+	if (bfs_debug(ctx, DEBUG_SEARCH, "eval_callback({\n")) {
 		fprintf(stderr, "\t.path = \"%s\",\n", ftwbuf->path);
 		fprintf(stderr, "\t.root = \"%s\",\n", ftwbuf->root);
 		fprintf(stderr, "\t.depth = %zu,\n", ftwbuf->depth);
 		fprintf(stderr, "\t.visit = %s,\n", dump_bftw_visit(ftwbuf->visit));
-		fprintf(stderr, "\t.typeflag = %s,\n", dump_bftw_typeflag(ftwbuf->typeflag));
+		fprintf(stderr, "\t.type = %s,\n", dump_bftw_type(ftwbuf->type));
 		fprintf(stderr, "\t.error = %d,\n", ftwbuf->error);
 		fprintf(stderr, "}) == %s\n", dump_bftw_action(state.action));
 	}
@@ -1234,7 +1377,7 @@ done:
 /**
  * Infer the number of open file descriptors we're allowed to have.
  */
-static int infer_fdlimit(const struct cmdline *cmdline) {
+static int infer_fdlimit(const struct bfs_ctx *ctx) {
 	int ret = 4096;
 
 	struct rlimit rl;
@@ -1245,7 +1388,7 @@ static int infer_fdlimit(const struct cmdline *cmdline) {
 	}
 
 	// 3 for std{in,out,err}
-	int nopen = 3 + cmdline->nopen_files;
+	int nopen = 3 + ctx->nfiles;
 
 	// Check /proc/self/fd for the current number of open fds, if possible
 	// (we may have inherited more than just the standard ones)
@@ -1266,8 +1409,8 @@ static int infer_fdlimit(const struct cmdline *cmdline) {
 	}
 
 	ret -= nopen;
-	ret -= cmdline->expr->persistent_fds;
-	ret -= cmdline->expr->ephemeral_fds;
+	ret -= ctx->expr->persistent_fds;
+	ret -= ctx->expr->ephemeral_fds;
 
 	// bftw() needs at least 2 available fds
 	if (ret < 2) {
@@ -1284,12 +1427,13 @@ static void dump_bftw_flags(enum bftw_flags flags) {
 	DEBUG_FLAG(flags, 0);
 	DEBUG_FLAG(flags, BFTW_STAT);
 	DEBUG_FLAG(flags, BFTW_RECOVER);
-	DEBUG_FLAG(flags, BFTW_DEPTH);
-	DEBUG_FLAG(flags, BFTW_COMFOLLOW);
-	DEBUG_FLAG(flags, BFTW_LOGICAL);
+	DEBUG_FLAG(flags, BFTW_POST_ORDER);
+	DEBUG_FLAG(flags, BFTW_FOLLOW_ROOTS);
+	DEBUG_FLAG(flags, BFTW_FOLLOW_ALL);
 	DEBUG_FLAG(flags, BFTW_DETECT_CYCLES);
-	DEBUG_FLAG(flags, BFTW_MOUNT);
-	DEBUG_FLAG(flags, BFTW_XDEV);
+	DEBUG_FLAG(flags, BFTW_SKIP_MOUNTS);
+	DEBUG_FLAG(flags, BFTW_PRUNE_MOUNTS);
+	DEBUG_FLAG(flags, BFTW_SORT);
 
 	assert(!flags);
 }
@@ -1302,49 +1446,53 @@ static const char *dump_bftw_strategy(enum bftw_strategy strategy) {
 		DUMP_BFTW_MAP(BFTW_BFS),
 		DUMP_BFTW_MAP(BFTW_DFS),
 		DUMP_BFTW_MAP(BFTW_IDS),
+		DUMP_BFTW_MAP(BFTW_EDS),
 	};
 	return strategies[strategy];
 }
 
-/**
- * Evaluate the command line.
- */
-int eval_cmdline(const struct cmdline *cmdline) {
-	if (!cmdline->expr) {
+int bfs_eval(const struct bfs_ctx *ctx) {
+	if (!ctx->expr) {
 		return EXIT_SUCCESS;
 	}
 
 	struct callback_args args = {
-		.cmdline = cmdline,
+		.ctx = ctx,
 		.ret = EXIT_SUCCESS,
 	};
 
+	if (ctx->status) {
+		args.bar = bfs_bar_show();
+		if (!args.bar) {
+			bfs_warning(ctx, "Couldn't show status bar: %m.\n\n");
+		}
+	}
+
 	struct trie seen;
-	if (cmdline->unique) {
+	if (ctx->unique) {
 		trie_init(&seen);
 		args.seen = &seen;
 	}
 
 	struct bftw_args bftw_args = {
-		.paths = cmdline->paths,
-		.npaths = darray_length(cmdline->paths),
-		.callback = cmdline_callback,
+		.paths = ctx->paths,
+		.npaths = darray_length(ctx->paths),
+		.callback = eval_callback,
 		.ptr = &args,
-		.nopenfd = infer_fdlimit(cmdline),
-		.flags = cmdline->flags,
-		.strategy = cmdline->strategy,
-		.mtab = cmdline->mtab,
+		.nopenfd = infer_fdlimit(ctx),
+		.flags = ctx->flags,
+		.strategy = ctx->strategy,
+		.mtab = bfs_ctx_mtab(ctx),
 	};
 
-	if (cmdline->debug & DEBUG_SEARCH) {
-		fprintf(stderr, "bftw({\n");
+	if (bfs_debug(ctx, DEBUG_SEARCH, "bftw({\n")) {
 		fprintf(stderr, "\t.paths = {\n");
 		for (size_t i = 0; i < bftw_args.npaths; ++i) {
 			fprintf(stderr, "\t\t\"%s\",\n", bftw_args.paths[i]);
 		}
 		fprintf(stderr, "\t},\n");
 		fprintf(stderr, "\t.npaths = %zu,\n", bftw_args.npaths);
-		fprintf(stderr, "\t.callback = cmdline_callback,\n");
+		fprintf(stderr, "\t.callback = eval_callback,\n");
 		fprintf(stderr, "\t.ptr = &args,\n");
 		fprintf(stderr, "\t.nopenfd = %d,\n", bftw_args.nopenfd);
 		fprintf(stderr, "\t.flags = ");
@@ -1352,7 +1500,7 @@ int eval_cmdline(const struct cmdline *cmdline) {
 		fprintf(stderr, ",\n\t.strategy = %s,\n", dump_bftw_strategy(bftw_args.strategy));
 		fprintf(stderr, "\t.mtab = ");
 		if (bftw_args.mtab) {
-			fprintf(stderr, "cmdline->mtab");
+			fprintf(stderr, "ctx->mtab");
 		} else {
 			fprintf(stderr, "NULL");
 		}
@@ -1361,20 +1509,20 @@ int eval_cmdline(const struct cmdline *cmdline) {
 
 	if (bftw(&bftw_args) != 0) {
 		args.ret = EXIT_FAILURE;
-		perror("bftw()");
+		bfs_perror(ctx, "bftw()");
 	}
 
-	if (eval_exec_finish(cmdline->expr, cmdline) != 0) {
+	if (eval_exec_finish(ctx->expr, ctx) != 0) {
 		args.ret = EXIT_FAILURE;
 	}
 
-	if (cmdline->debug & DEBUG_RATES) {
-		dump_cmdline(cmdline, true);
-	}
+	bfs_ctx_dump(ctx, DEBUG_RATES);
 
-	if (cmdline->unique) {
+	if (ctx->unique) {
 		trie_destroy(&seen);
 	}
 
+	bfs_bar_hide(args.bar);
+
 	return args.ret;
 }
