@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2017-2019 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2017-2020 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -39,14 +39,18 @@
  * effects are reachable at all, and skipping the traversal if not.
  */
 
-#include "cmdline.h"
+#include "opt.h"
 #include "color.h"
+#include "ctx.h"
+#include "diag.h"
 #include "eval.h"
 #include "expr.h"
+#include "pwcache.h"
 #include <assert.h>
 #include <limits.h>
 #include <stdarg.h>
 #include <stdio.h>
+#include <unistd.h>
 
 static char *fake_and_arg = "-a";
 static char *fake_or_arg = "-o";
@@ -116,7 +120,7 @@ static void range_union(struct range *result, const struct range *lhs, const str
 }
 
 /** Check if a range contains no values. */
-static bool range_impossible(const struct range *range) {
+static bool range_is_impossible(const struct range *range) {
 	return range->min > range->max;
 }
 
@@ -143,7 +147,73 @@ enum range_type {
 	/** User ID. */
 	UID_RANGE,
 	/** The number of range_types. */
-	MAX_RANGE,
+	RANGE_TYPES,
+};
+
+/**
+ * A possibly-known value of a predicate.
+ */
+enum known_pred {
+	/** The state is impossible to reach. */
+	PRED_IMPOSSIBLE = -2,
+	/** The value of the predicate is not known. */
+	PRED_UNKNOWN = -1,
+	/** The predicate is known to be false. */
+	PRED_FALSE = false,
+	/** The predicate is known to be true. */
+	PRED_TRUE = true,
+};
+
+/** Make a predicate known. */
+static void constrain_pred(enum known_pred *pred, bool value) {
+	if (*pred == PRED_UNKNOWN) {
+		*pred = value;
+	} else if (*pred == !value) {
+		*pred = PRED_IMPOSSIBLE;
+	}
+}
+
+/** Compute the union of two known predicates. */
+static enum known_pred pred_union(enum known_pred lhs, enum known_pred rhs) {
+	if (lhs == PRED_IMPOSSIBLE) {
+		return rhs;
+	} else if (rhs == PRED_IMPOSSIBLE) {
+		return lhs;
+	} else if (lhs == rhs) {
+		return lhs;
+	} else {
+		return PRED_UNKNOWN;
+	}
+}
+
+/**
+ * Types of predicates we track.
+ */
+enum pred_type {
+	/** -readable */
+	READABLE_PRED,
+	/** -writable */
+	WRITABLE_PRED,
+	/** -executable */
+	EXECUTABLE_PRED,
+	/** -acl */
+	ACL_PRED,
+	/** -capable */
+	CAPABLE_PRED,
+	/** -empty */
+	EMPTY_PRED,
+	/** -hidden */
+	HIDDEN_PRED,
+	/** -nogroup */
+	NOGROUP_PRED,
+	/** -nouser */
+	NOUSER_PRED,
+	/** -sparse */
+	SPARSE_PRED,
+	/** -xattr */
+	XATTR_PRED,
+	/** The number of pred_types. */
+	PRED_TYPES,
 };
 
 /**
@@ -151,30 +221,41 @@ enum range_type {
  */
 struct opt_facts {
 	/** The value ranges we track. */
-	struct range ranges[MAX_RANGE];
+	struct range ranges[RANGE_TYPES];
+
+	/** The predicates we track. */
+	enum known_pred preds[PRED_TYPES];
 
 	/** Bitmask of possible file types. */
-	enum bftw_typeflag types;
+	unsigned int types;
 	/** Bitmask of possible link target types. */
-	enum bftw_typeflag xtypes;
+	unsigned int xtypes;
 };
 
 /** Initialize some data flow facts. */
 static void facts_init(struct opt_facts *facts) {
-	for (int i = 0; i < MAX_RANGE; ++i) {
-		struct range *range = facts->ranges + i;
+	for (int i = 0; i < RANGE_TYPES; ++i) {
+		struct range *range = &facts->ranges[i];
 		range->min = 0; // All ranges we currently track are non-negative
 		range->max = LLONG_MAX;
 	}
 
+	for (int i = 0; i < PRED_TYPES; ++i) {
+		facts->preds[i] = PRED_UNKNOWN;
+	}
+
 	facts->types = ~0;
 	facts->xtypes = ~0;
 }
 
 /** Compute the union of two fact sets. */
 static void facts_union(struct opt_facts *result, const struct opt_facts *lhs, const struct opt_facts *rhs) {
-	for (int i = 0; i < MAX_RANGE; ++i) {
-		range_union(result->ranges + i, lhs->ranges + i, rhs->ranges + i);
+	for (int i = 0; i < RANGE_TYPES; ++i) {
+		range_union(&result->ranges[i], &lhs->ranges[i], &rhs->ranges[i]);
+	}
+
+	for (int i = 0; i < PRED_TYPES; ++i) {
+		result->preds[i] = pred_union(lhs->preds[i], rhs->preds[i]);
 	}
 
 	result->types = lhs->types | rhs->types;
@@ -182,9 +263,15 @@ static void facts_union(struct opt_facts *result, const struct opt_facts *lhs, c
 }
 
 /** Determine whether a fact set is impossible. */
-static bool facts_impossible(const struct opt_facts *facts) {
-	for (int i = 0; i < MAX_RANGE; ++i) {
-		if (range_impossible(facts->ranges + i)) {
+static bool facts_are_impossible(const struct opt_facts *facts) {
+	for (int i = 0; i < RANGE_TYPES; ++i) {
+		if (range_is_impossible(&facts->ranges[i])) {
+			return true;
+		}
+	}
+
+	for (int i = 0; i < PRED_TYPES; ++i) {
+		if (facts->preds[i] == PRED_IMPOSSIBLE) {
 			return true;
 		}
 	}
@@ -198,8 +285,12 @@ static bool facts_impossible(const struct opt_facts *facts) {
 
 /** Set some facts to be impossible. */
 static void set_facts_impossible(struct opt_facts *facts) {
-	for (int i = 0; i < MAX_RANGE; ++i) {
-		set_range_impossible(facts->ranges + i);
+	for (int i = 0; i < RANGE_TYPES; ++i) {
+		set_range_impossible(&facts->ranges[i]);
+	}
+
+	for (int i = 0; i < PRED_TYPES; ++i) {
+		facts->preds[i] = PRED_IMPOSSIBLE;
 	}
 
 	facts->types = 0;
@@ -210,8 +301,8 @@ static void set_facts_impossible(struct opt_facts *facts) {
  * Optimizer state.
  */
 struct opt_state {
-	/** The command line we're optimizing. */
-	const struct cmdline *cmdline;
+	/** The context we're optimizing. */
+	const struct bfs_ctx *ctx;
 
 	/** Data flow facts before this expression is evaluated. */
 	struct opt_facts facts;
@@ -224,41 +315,19 @@ struct opt_state {
 };
 
 /** Log an optimization. */
-static void debug_opt(const struct opt_state *state, const char *format, ...) {
-	if (!(state->cmdline->debug & DEBUG_OPT)) {
-		return;
+BFS_FORMATTER(3, 4)
+static bool debug_opt(const struct opt_state *state, int level, const char *format, ...) {
+	assert(state->ctx->optlevel >= level);
+
+	if (bfs_debug(state->ctx, DEBUG_OPT, "${cyn}-O%d${rs}: ", level)) {
+		va_list args;
+		va_start(args, format);
+		cvfprintf(state->ctx->cerr, format, args);
+		va_end(args);
+		return true;
+	} else {
+		return false;
 	}
-
-	CFILE *cerr = state->cmdline->cerr;
-
-	va_list args;
-	va_start(args, format);
-
-	for (const char *i = format; *i != '\0'; ++i) {
-		if (*i == '%') {
-			switch (*++i) {
-			case 'd':
-				fprintf(cerr->file, "%d", va_arg(args, int));
-				break;
-
-			case 'e':
-				dump_expr(cerr, va_arg(args, const struct expr *), false);
-				break;
-
-			case 'g':
-				cfprintf(cerr, "${ylw}%g${rs}", va_arg(args, double));
-				break;
-
-			default:
-				assert(false);
-				break;
-			}
-		} else {
-			fputc(*i, stderr);
-		}
-	}
-
-	va_end(args);
 }
 
 /** Extract a child expression, freeing the outer expression. */
@@ -295,7 +364,7 @@ static struct expr *optimize_or_expr(const struct opt_state *state, struct expr
  * Apply De Morgan's laws.
  */
 static struct expr *de_morgan(const struct opt_state *state, struct expr *expr, char **argv) {
-	debug_opt(state, "-O1: De Morgan's laws: %e ", expr);
+	bool debug = debug_opt(state, 1, "De Morgan's laws: %pe ", expr);
 
 	struct expr *parent = negate_expr(expr, argv);
 	if (!parent) {
@@ -324,7 +393,9 @@ static struct expr *de_morgan(const struct opt_state *state, struct expr *expr,
 		return NULL;
 	}
 
-	debug_opt(state, "<==> %e\n", parent);
+	if (debug) {
+		cfprintf(state->ctx->cerr, "<==> %pe\n", parent);
+	}
 
 	if (expr->lhs->eval == eval_not) {
 		expr->lhs = optimize_not_expr(state, expr->lhs);
@@ -369,21 +440,21 @@ static struct expr *optimize_not_expr(const struct opt_state *state, struct expr
 
 	struct expr *rhs = expr->rhs;
 
-	int optlevel = state->cmdline->optlevel;
+	int optlevel = state->ctx->optlevel;
 	if (optlevel >= 1) {
 		if (rhs == &expr_true) {
-			debug_opt(state, "-O1: constant propagation: %e <==> %e\n", expr, &expr_false);
+			debug_opt(state, 1, "constant propagation: %pe <==> %pe\n", expr, &expr_false);
 			free_expr(expr);
 			return &expr_false;
 		} else if (rhs == &expr_false) {
-			debug_opt(state, "-O1: constant propagation: %e <==> %e\n", expr, &expr_true);
+			debug_opt(state, 1, "constant propagation: %pe <==> %pe\n", expr, &expr_true);
 			free_expr(expr);
 			return &expr_true;
 		} else if (rhs->eval == eval_not) {
-			debug_opt(state, "-O1: double negation: %e <==> %e\n", expr, rhs->rhs);
+			debug_opt(state, 1, "double negation: %pe <==> %pe\n", expr, rhs->rhs);
 			return extract_child_expr(expr, &rhs->rhs);
 		} else if (expr_never_returns(rhs)) {
-			debug_opt(state, "-O1: reachability: %e <==> %e\n", expr, rhs);
+			debug_opt(state, 1, "reachability: %pe <==> %pe\n", expr, rhs);
 			return extract_child_expr(expr, &expr->rhs);
 		} else if ((rhs->eval == eval_and || rhs->eval == eval_or)
 			   && (rhs->lhs->eval == eval_not || rhs->rhs->eval == eval_not)) {
@@ -425,27 +496,28 @@ static struct expr *optimize_and_expr(const struct opt_state *state, struct expr
 	struct expr *lhs = expr->lhs;
 	struct expr *rhs = expr->rhs;
 
-	int optlevel = state->cmdline->optlevel;
+	const struct bfs_ctx *ctx = state->ctx;
+	int optlevel = ctx->optlevel;
 	if (optlevel >= 1) {
 		if (lhs == &expr_true) {
-			debug_opt(state, "-O1: conjunction elimination: %e <==> %e\n", expr, rhs);
+			debug_opt(state, 1, "conjunction elimination: %pe <==> %pe\n", expr, rhs);
 			return extract_child_expr(expr, &expr->rhs);
 		} else if (rhs == &expr_true) {
-			debug_opt(state, "-O1: conjunction elimination: %e <==> %e\n", expr, lhs);
+			debug_opt(state, 1, "conjunction elimination: %pe <==> %pe\n", expr, lhs);
 			return extract_child_expr(expr, &expr->lhs);
 		} else if (lhs->always_false) {
-			debug_opt(state, "-O1: short-circuit: %e <==> %e\n", expr, lhs);
+			debug_opt(state, 1, "short-circuit: %pe <==> %pe\n", expr, lhs);
 			return extract_child_expr(expr, &expr->lhs);
 		} else if (lhs->always_true && rhs == &expr_false) {
-			debug_opt(state, "-O1: strength reduction: %e <==> ", expr);
+			bool debug = debug_opt(state, 1, "strength reduction: %pe <==> ", expr);
 			struct expr *ret = extract_child_expr(expr, &expr->lhs);
 			ret = negate_expr(ret, &fake_not_arg);
-			if (ret) {
-				debug_opt(state, "%e\n", ret);
+			if (debug && ret) {
+				cfprintf(ctx->cerr, "%pe\n", ret);
 			}
 			return ret;
 		} else if (optlevel >= 2 && lhs->pure && rhs == &expr_false) {
-			debug_opt(state, "-O2: purity: %e <==> %e\n", expr, rhs);
+			debug_opt(state, 2, "purity: %pe <==> %pe\n", expr, rhs);
 			return extract_child_expr(expr, &expr->rhs);
 		} else if (lhs->eval == eval_not && rhs->eval == eval_not) {
 			return de_morgan(state, expr, expr->lhs->argv);
@@ -493,27 +565,28 @@ static struct expr *optimize_or_expr(const struct opt_state *state, struct expr
 	struct expr *lhs = expr->lhs;
 	struct expr *rhs = expr->rhs;
 
-	int optlevel = state->cmdline->optlevel;
+	const struct bfs_ctx *ctx = state->ctx;
+	int optlevel = ctx->optlevel;
 	if (optlevel >= 1) {
 		if (lhs->always_true) {
-			debug_opt(state, "-O1: short-circuit: %e <==> %e\n", expr, lhs);
+			debug_opt(state, 1, "short-circuit: %pe <==> %pe\n", expr, lhs);
 			return extract_child_expr(expr, &expr->lhs);
 		} else if (lhs == &expr_false) {
-			debug_opt(state, "-O1: disjunctive syllogism: %e <==> %e\n", expr, rhs);
+			debug_opt(state, 1, "disjunctive syllogism: %pe <==> %pe\n", expr, rhs);
 			return extract_child_expr(expr, &expr->rhs);
 		} else if (rhs == &expr_false) {
-			debug_opt(state, "-O1: disjunctive syllogism: %e <==> %e\n", expr, lhs);
+			debug_opt(state, 1, "disjunctive syllogism: %pe <==> %pe\n", expr, lhs);
 			return extract_child_expr(expr, &expr->lhs);
 		} else if (lhs->always_false && rhs == &expr_true) {
-			debug_opt(state, "-O1: strength reduction: %e <==> ", expr);
+			bool debug = debug_opt(state, 1, "strength reduction: %pe <==> ", expr);
 			struct expr *ret = extract_child_expr(expr, &expr->lhs);
 			ret = negate_expr(ret, &fake_not_arg);
-			if (ret) {
-				debug_opt(state, "%e\n", ret);
+			if (debug && ret) {
+				cfprintf(ctx->cerr, "%pe\n", ret);
 			}
 			return ret;
 		} else if (optlevel >= 2 && lhs->pure && rhs == &expr_true) {
-			debug_opt(state, "-O2: purity: %e <==> %e\n", expr, rhs);
+			debug_opt(state, 2, "purity: %pe <==> %pe\n", expr, rhs);
 			return extract_child_expr(expr, &expr->rhs);
 		} else if (lhs->eval == eval_not && rhs->eval == eval_not) {
 			return de_morgan(state, expr, expr->lhs->argv);
@@ -556,17 +629,17 @@ fail:
 
 /** Optimize an expression in an ignored-result context. */
 static struct expr *ignore_result(const struct opt_state *state, struct expr *expr) {
-	int optlevel = state->cmdline->optlevel;
+	int optlevel = state->ctx->optlevel;
 
 	if (optlevel >= 1) {
 		while (true) {
 			if (expr->eval == eval_not) {
-				debug_opt(state, "-O1: ignored result: %e --> %e\n", expr, expr->rhs);
+				debug_opt(state, 1, "ignored result: %pe --> %pe\n", expr, expr->rhs);
 				expr = extract_child_expr(expr, &expr->rhs);
 			} else if (optlevel >= 2
 			           && (expr->eval == eval_and || expr->eval == eval_or || expr->eval == eval_comma)
 			           && expr->rhs->pure) {
-				debug_opt(state, "-O2: ignored result: %e --> %e\n", expr, expr->lhs);
+				debug_opt(state, 2, "ignored result: %pe --> %pe\n", expr, expr->lhs);
 				expr = extract_child_expr(expr, &expr->lhs);
 			} else {
 				break;
@@ -574,7 +647,7 @@ static struct expr *ignore_result(const struct opt_state *state, struct expr *ex
 		}
 
 		if (optlevel >= 2 && expr->pure && expr != &expr_false) {
-			debug_opt(state, "-O2: ignored result: %e --> %e\n", expr, &expr_false);
+			debug_opt(state, 2, "ignored result: %pe --> %pe\n", expr, &expr_false);
 			free_expr(expr);
 			expr = &expr_false;
 		}
@@ -590,19 +663,19 @@ static struct expr *optimize_comma_expr(const struct opt_state *state, struct ex
 	struct expr *lhs = expr->lhs;
 	struct expr *rhs = expr->rhs;
 
-	int optlevel = state->cmdline->optlevel;
+	int optlevel = state->ctx->optlevel;
 	if (optlevel >= 1) {
 		lhs = expr->lhs = ignore_result(state, lhs);
 
 		if (expr_never_returns(lhs)) {
-			debug_opt(state, "-O1: reachability: %e <==> %e\n", expr, lhs);
+			debug_opt(state, 1, "reachability: %pe <==> %pe\n", expr, lhs);
 			return extract_child_expr(expr, &expr->lhs);
 		} else if ((lhs->always_true && rhs == &expr_true)
 			   || (lhs->always_false && rhs == &expr_false)) {
-			debug_opt(state, "-O1: redundancy elimination: %e <==> %e\n", expr, lhs);
+			debug_opt(state, 1, "redundancy elimination: %pe <==> %pe\n", expr, lhs);
 			return extract_child_expr(expr, &expr->lhs);
 		} else if (optlevel >= 2 && lhs->pure) {
-			debug_opt(state, "-O2: purity: %e <==> %e\n", expr, rhs);
+			debug_opt(state, 2, "purity: %pe <==> %pe\n", expr, rhs);
 			return extract_child_expr(expr, &expr->rhs);
 		}
 	}
@@ -638,10 +711,29 @@ fail:
 	return NULL;
 }
 
-/** Infer data flow facts about an icmp-style ([+-]N) expression */
+/** Infer data flow facts about a predicate. */
+static void infer_pred_facts(struct opt_state *state, enum pred_type pred) {
+	constrain_pred(&state->facts_when_true.preds[pred], true);
+	constrain_pred(&state->facts_when_false.preds[pred], false);
+}
+
+/** Infer data flow facts about an -{execut,read,writ}able expression. */
+static void infer_access_facts(struct opt_state *state, const struct expr *expr) {
+	if (expr->idata & R_OK) {
+		infer_pred_facts(state, READABLE_PRED);
+	}
+	if (expr->idata & W_OK) {
+		infer_pred_facts(state, WRITABLE_PRED);
+	}
+	if (expr->idata & X_OK) {
+		infer_pred_facts(state, EXECUTABLE_PRED);
+	}
+}
+
+/** Infer data flow facts about an icmp-style ([+-]N) expression. */
 static void infer_icmp_facts(struct opt_state *state, const struct expr *expr, enum range_type type) {
-	struct range *range_when_true = state->facts_when_true.ranges + type;
-	struct range *range_when_false = state->facts_when_false.ranges + type;
+	struct range *range_when_true = &state->facts_when_true.ranges[type];
+	struct range *range_when_false =& state->facts_when_false.ranges[type];
 	long long value = expr->idata;
 
 	switch (expr->cmp_flag) {
@@ -665,9 +757,35 @@ static void infer_icmp_facts(struct opt_state *state, const struct expr *expr, e
 	}
 }
 
+/** Infer data flow facts about a -gid expression. */
+static void infer_gid_facts(struct opt_state *state, const struct expr *expr) {
+	infer_icmp_facts(state, expr, GID_RANGE);
+
+	const struct bfs_groups *groups = bfs_ctx_groups(state->ctx);
+	struct range *range = &state->facts_when_true.ranges[GID_RANGE];
+	if (groups && range->min == range->max) {
+		gid_t gid = range->min;
+		bool nogroup = !bfs_getgrgid(groups, gid);
+		constrain_pred(&state->facts_when_true.preds[NOGROUP_PRED], nogroup);
+	}
+}
+
+/** Infer data flow facts about a -uid expression. */
+static void infer_uid_facts(struct opt_state *state, const struct expr *expr) {
+	infer_icmp_facts(state, expr, UID_RANGE);
+
+	const struct bfs_users *users = bfs_ctx_users(state->ctx);
+	struct range *range = &state->facts_when_true.ranges[UID_RANGE];
+	if (users && range->min == range->max) {
+		uid_t uid = range->min;
+		bool nouser = !bfs_getpwuid(users, uid);
+		constrain_pred(&state->facts_when_true.preds[NOUSER_PRED], nouser);
+	}
+}
+
 /** Infer data flow facts about a -samefile expression. */
 static void infer_samefile_facts(struct opt_state *state, const struct expr *expr) {
-	struct range *range_when_true = state->facts_when_true.ranges + INUM_RANGE;
+	struct range *range_when_true = &state->facts_when_true.ranges[INUM_RANGE];
 	constrain_min(range_when_true, expr->ino);
 	constrain_max(range_when_true, expr->ino);
 }
@@ -685,25 +803,56 @@ static void infer_xtype_facts(struct opt_state *state, const struct expr *expr)
 }
 
 static struct expr *optimize_expr_recursive(struct opt_state *state, struct expr *expr) {
+	int optlevel = state->ctx->optlevel;
+
 	state->facts_when_true = state->facts;
 	state->facts_when_false = state->facts;
 
-	if (expr->eval == eval_depth) {
+	if (optlevel >= 2 && facts_are_impossible(&state->facts)) {
+		debug_opt(state, 2, "reachability: %pe --> %pe\n", expr, &expr_false);
+		free_expr(expr);
+		expr = &expr_false;
+		goto done;
+	}
+
+	if (!expr->rhs && !expr->pure) {
+		facts_union(state->facts_when_impure, state->facts_when_impure, &state->facts);
+	}
+
+	if (expr->eval == eval_access) {
+		infer_access_facts(state, expr);
+	} else if (expr->eval == eval_acl) {
+		infer_pred_facts(state, ACL_PRED);
+	} else if (expr->eval == eval_capable) {
+		infer_pred_facts(state, CAPABLE_PRED);
+	} else if (expr->eval == eval_depth) {
 		infer_icmp_facts(state, expr, DEPTH_RANGE);
+	} else if (expr->eval == eval_empty) {
+		infer_pred_facts(state, EMPTY_PRED);
 	} else if (expr->eval == eval_gid) {
-		infer_icmp_facts(state, expr, GID_RANGE);
+		infer_gid_facts(state, expr);
+	} else if (expr->eval == eval_hidden) {
+		infer_pred_facts(state, HIDDEN_PRED);
 	} else if (expr->eval == eval_inum) {
 		infer_icmp_facts(state, expr, INUM_RANGE);
 	} else if (expr->eval == eval_links) {
 		infer_icmp_facts(state, expr, LINKS_RANGE);
+	} else if (expr->eval == eval_nogroup) {
+		infer_pred_facts(state, NOGROUP_PRED);
+	} else if (expr->eval == eval_nouser) {
+		infer_pred_facts(state, NOUSER_PRED);
 	} else if (expr->eval == eval_samefile) {
 		infer_samefile_facts(state, expr);
 	} else if (expr->eval == eval_size) {
 		infer_icmp_facts(state, expr, SIZE_RANGE);
+	} else if (expr->eval == eval_sparse) {
+		infer_pred_facts(state, SPARSE_PRED);
 	} else if (expr->eval == eval_type) {
 		infer_type_facts(state, expr);
 	} else if (expr->eval == eval_uid) {
-		infer_icmp_facts(state, expr, UID_RANGE);
+		infer_uid_facts(state, expr);
+	} else if (expr->eval == eval_xattr) {
+		infer_pred_facts(state, XATTR_PRED);
 	} else if (expr->eval == eval_xtype) {
 		infer_xtype_facts(state, expr);
 	} else if (expr->eval == eval_not) {
@@ -714,8 +863,6 @@ static struct expr *optimize_expr_recursive(struct opt_state *state, struct expr
 		expr = optimize_or_expr_recursive(state, expr);
 	} else if (expr->eval == eval_comma) {
 		expr = optimize_comma_expr_recursive(state, expr);
-	} else if (!expr->pure) {
-		facts_union(state->facts_when_impure, state->facts_when_impure, &state->facts);
 	}
 
 	if (!expr) {
@@ -742,22 +889,22 @@ static struct expr *optimize_expr_recursive(struct opt_state *state, struct expr
 		set_facts_impossible(&state->facts_when_true);
 	}
 
-	if (state->cmdline->optlevel < 2 || expr == &expr_true || expr == &expr_false) {
+	if (optlevel < 2 || expr == &expr_true || expr == &expr_false) {
 		goto done;
 	}
 
-	if (facts_impossible(&state->facts_when_true)) {
+	if (facts_are_impossible(&state->facts_when_true)) {
 		if (expr->pure) {
-			debug_opt(state, "-O2: data flow: %e --> %e\n", expr, &expr_false);
+			debug_opt(state, 2, "data flow: %pe --> %pe\n", expr, &expr_false);
 			free_expr(expr);
 			expr = &expr_false;
 		} else {
 			expr->always_false = true;
 			expr->probability = 0.0;
 		}
-	} else if (facts_impossible(&state->facts_when_false)) {
+	} else if (facts_are_impossible(&state->facts_when_false)) {
 		if (expr->pure) {
-			debug_opt(state, "-O2: data flow: %e --> %e\n", expr, &expr_true);
+			debug_opt(state, 2, "data flow: %pe --> %pe\n", expr, &expr_true);
 			free_expr(expr);
 			expr = &expr_true;
 		} else {
@@ -773,11 +920,13 @@ done:
 /** Swap the children of a binary expression if it would reduce the cost. */
 static bool reorder_expr(const struct opt_state *state, struct expr *expr, double swapped_cost) {
 	if (swapped_cost < expr->cost) {
-		debug_opt(state, "-O3: cost: %e", expr);
+		bool debug = debug_opt(state, 3, "cost: %pe <==> ", expr);
 		struct expr *lhs = expr->lhs;
 		expr->lhs = expr->rhs;
 		expr->rhs = lhs;
-		debug_opt(state, " <==> %e (~%g --> ~%g)\n", expr, expr->cost, swapped_cost);
+		if (debug) {
+			cfprintf(state->ctx->cerr, "%pe (~${ylw}%g${rs} --> ~${ylw}%g${rs})\n", expr, expr->cost, swapped_cost);
+		}
 		expr->cost = swapped_cost;
 		return true;
 	} else {
@@ -816,56 +965,84 @@ static bool reorder_expr_recursive(const struct opt_state *state, struct expr *e
 	return ret;
 }
 
-int optimize_cmdline(struct cmdline *cmdline) {
+/**
+ * Optimize a top-level expression.
+ */
+static struct expr *optimize_expr(struct opt_state *state, struct expr *expr) {
+	struct opt_facts saved_impure = *state->facts_when_impure;
+
+	expr = optimize_expr_recursive(state, expr);
+	if (!expr) {
+		return NULL;
+	}
+
+	if (state->ctx->optlevel >= 3 && reorder_expr_recursive(state, expr)) {
+		// Re-do optimizations to account for the new ordering
+		*state->facts_when_impure = saved_impure;
+		expr = optimize_expr_recursive(state, expr);
+		if (!expr) {
+			return NULL;
+		}
+	}
+
+	return expr;
+}
+
+int bfs_optimize(struct bfs_ctx *ctx) {
+	bfs_ctx_dump(ctx, DEBUG_OPT);
+
 	struct opt_facts facts_when_impure;
 	set_facts_impossible(&facts_when_impure);
 
 	struct opt_state state = {
-		.cmdline = cmdline,
+		.ctx = ctx,
 		.facts_when_impure = &facts_when_impure,
 	};
 	facts_init(&state.facts);
 
-	struct range *depth = state.facts.ranges + DEPTH_RANGE;
-	depth->min = cmdline->mindepth;
-	depth->max = cmdline->maxdepth;
-
-	int optlevel = cmdline->optlevel;
-
-	cmdline->expr = optimize_expr_recursive(&state, cmdline->expr);
-	if (!cmdline->expr) {
+	ctx->exclude = optimize_expr(&state, ctx->exclude);
+	if (!ctx->exclude) {
 		return -1;
 	}
 
-	if (optlevel >= 3 && reorder_expr_recursive(&state, cmdline->expr)) {
-		// Re-do optimizations to account for the new ordering
-		set_facts_impossible(&facts_when_impure);
-		cmdline->expr = optimize_expr_recursive(&state, cmdline->expr);
-		if (!cmdline->expr) {
-			return -1;
-		}
+	// Only non-excluded files are evaluated
+	state.facts = state.facts_when_false;
+
+	struct range *depth = &state.facts.ranges[DEPTH_RANGE];
+	constrain_min(depth, ctx->mindepth);
+	constrain_max(depth, ctx->maxdepth);
+
+	ctx->expr = optimize_expr(&state, ctx->expr);
+	if (!ctx->expr) {
+		return -1;
 	}
 
-	cmdline->expr = ignore_result(&state, cmdline->expr);
+	ctx->expr = ignore_result(&state, ctx->expr);
 
-	const struct range *depth_when_impure = facts_when_impure.ranges + DEPTH_RANGE;
+	if (facts_are_impossible(&facts_when_impure)) {
+		bfs_warning(ctx, "This command won't do anything.\n\n");
+	}
+
+	const struct range *depth_when_impure = &facts_when_impure.ranges[DEPTH_RANGE];
 	long long mindepth = depth_when_impure->min;
 	long long maxdepth = depth_when_impure->max;
 
-	if (optlevel >= 2 && mindepth > cmdline->mindepth) {
+	int optlevel = ctx->optlevel;
+
+	if (optlevel >= 2 && mindepth > ctx->mindepth) {
 		if (mindepth > INT_MAX) {
 			mindepth = INT_MAX;
 		}
-		cmdline->mindepth = mindepth;
-		debug_opt(&state, "-O2: data flow: mindepth --> %d\n", cmdline->mindepth);
+		ctx->mindepth = mindepth;
+		debug_opt(&state, 2, "data flow: mindepth --> %d\n", ctx->mindepth);
 	}
 
-	if (optlevel >= 4 && maxdepth < cmdline->maxdepth) {
+	if (optlevel >= 4 && maxdepth < ctx->maxdepth) {
 		if (maxdepth < INT_MIN) {
 			maxdepth = INT_MIN;
 		}
-		cmdline->maxdepth = maxdepth;
-		debug_opt(&state, "-O4: data flow: maxdepth --> %d\n", cmdline->maxdepth);
+		ctx->maxdepth = maxdepth;
+		debug_opt(&state, 4, "data flow: maxdepth --> %d\n", ctx->maxdepth);
 	}
 
 	return 0;
