@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2017-2019 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2017-2020 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -15,8 +15,8 @@
  ****************************************************************************/
 
 #include "printf.h"
-#include "cmdline.h"
 #include "color.h"
+#include "ctx.h"
 #include "diag.h"
 #include "dstring.h"
 #include "expr.h"
@@ -73,7 +73,7 @@ static int bfs_printf_flush(FILE *file, const struct bfs_printf *directive, cons
 #define BFS_PRINTF_BUF(buf, format, ...)				\
 	char buf[256];							\
 	int ret = snprintf(buf, sizeof(buf), format, __VA_ARGS__);	\
-	assert(ret >= 0 && ret < sizeof(buf));				\
+	assert(ret >= 0 && (size_t)ret < sizeof(buf));			\
 	(void)ret
 
 /** %a, %c, %t: ctime() */
@@ -172,7 +172,7 @@ static int bfs_printf_strftime(FILE *file, const struct bfs_printf *directive, c
 		break;
 	}
 
-	assert(ret >= 0 && ret < sizeof(buf));
+	assert(ret >= 0 && (size_t)ret < sizeof(buf));
 	(void)ret;
 
 	return fprintf(file, directive->str, buf);
@@ -306,17 +306,21 @@ static int bfs_printf_k(FILE *file, const struct bfs_printf *directive, const st
 
 /** %l: link target */
 static int bfs_printf_l(FILE *file, const struct bfs_printf *directive, const struct BFTW *ftwbuf) {
-	if (ftwbuf->typeflag != BFTW_LNK) {
-		return 0;
-	}
+	char *buf = NULL;
+	const char *target = "";
 
-	char *target = xreadlinkat(ftwbuf->at_fd, ftwbuf->at_path, 0);
-	if (!target) {
-		return -1;
+	if (ftwbuf->type == BFTW_LNK) {
+		const struct bfs_stat *statbuf = bftw_cached_stat(ftwbuf, BFS_STAT_NOFOLLOW);
+		size_t len = statbuf ? statbuf->size : 0;
+
+		target = buf = xreadlinkat(ftwbuf->at_fd, ftwbuf->at_path, len);
+		if (!target) {
+			return -1;
+		}
 	}
 
 	int ret = fprintf(file, directive->str, target);
-	free(target);
+	free(buf);
 	return ret;
 }
 
@@ -338,7 +342,7 @@ static int bfs_printf_M(FILE *file, const struct bfs_printf *directive, const st
 	}
 
 	char buf[11];
-	format_mode(statbuf->mode, buf);
+	xstrmode(statbuf->mode, buf);
 	return fprintf(file, directive->str, buf);
 }
 
@@ -421,8 +425,8 @@ static int bfs_printf_u(FILE *file, const struct bfs_printf *directive, const st
 	return fprintf(file, directive->str, pwd->pw_name);
 }
 
-static const char *bfs_printf_type(enum bftw_typeflag typeflag) {
-	switch (typeflag) {
+static const char *bfs_printf_type(enum bftw_type type) {
+	switch (type) {
 	case BFTW_BLK:
 		return "b";
 	case BFTW_CHR:
@@ -446,7 +450,7 @@ static const char *bfs_printf_type(enum bftw_typeflag typeflag) {
 
 /** %y: type */
 static int bfs_printf_y(FILE *file, const struct bfs_printf *directive, const struct BFTW *ftwbuf) {
-	const char *type = bfs_printf_type(ftwbuf->typeflag);
+	const char *type = bfs_printf_type(ftwbuf->type);
 	return fprintf(file, directive->str, type);
 }
 
@@ -454,7 +458,7 @@ static int bfs_printf_y(FILE *file, const struct bfs_printf *directive, const st
 static int bfs_printf_Y(FILE *file, const struct bfs_printf *directive, const struct BFTW *ftwbuf) {
 	int error = 0;
 
-	if (ftwbuf->typeflag != BFTW_LNK) {
+	if (ftwbuf->type != BFTW_LNK) {
 		return bfs_printf_y(file, directive, ftwbuf);
 	}
 
@@ -462,7 +466,7 @@ static int bfs_printf_Y(FILE *file, const struct bfs_printf *directive, const st
 
 	const struct bfs_stat *statbuf = bftw_stat(ftwbuf, BFS_STAT_FOLLOW);
 	if (statbuf) {
-		type = bfs_printf_type(bftw_mode_typeflag(statbuf->mode));
+		type = bfs_printf_type(bftw_mode_to_type(statbuf->mode));
 	} else {
 		switch (errno) {
 		case ELOOP:
@@ -500,17 +504,17 @@ static void free_directive(struct bfs_printf *directive) {
 /**
  * Create a new printf directive.
  */
-static struct bfs_printf *new_directive(bfs_printf_fn *fn) {
+static struct bfs_printf *new_directive(const struct bfs_ctx *ctx, bfs_printf_fn *fn) {
 	struct bfs_printf *directive = malloc(sizeof(*directive));
 	if (!directive) {
-		perror("malloc()");
+		bfs_perror(ctx, "malloc()");
 		goto error;
 	}
 
 	directive->fn = fn;
 	directive->str = dstralloc(2);
 	if (!directive->str) {
-		perror("dstralloc()");
+		bfs_perror(ctx, "dstralloc()");
 		goto error;
 	}
 	directive->stat_field = 0;
@@ -546,11 +550,11 @@ static struct bfs_printf **append_literal(struct bfs_printf **tail, struct bfs_p
 	}
 }
 
-struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline) {
+struct bfs_printf *bfs_printf_parse(const struct bfs_ctx *ctx, const char *format) {
 	struct bfs_printf *head = NULL;
 	struct bfs_printf **tail = &head;
 
-	struct bfs_printf *literal = new_directive(bfs_printf_literal);
+	struct bfs_printf *literal = new_directive(ctx, bfs_printf_literal);
 	if (!literal) {
 		goto error;
 	}
@@ -583,7 +587,7 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 
 			case 'c':
 				tail = append_literal(tail, &literal);
-				struct bfs_printf *directive = new_directive(bfs_printf_flush);
+				struct bfs_printf *directive = new_directive(ctx, bfs_printf_flush);
 				if (!directive) {
 					goto error;
 				}
@@ -591,11 +595,11 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 				goto done;
 
 			case '\0':
-				bfs_error(cmdline, "'%s': Incomplete escape sequence '\\'.\n", format);
+				bfs_error(ctx, "'%s': Incomplete escape sequence '\\'.\n", format);
 				goto error;
 
 			default:
-				bfs_error(cmdline, "'%s': Unrecognized escape sequence '\\%c'.\n", format, c);
+				bfs_error(ctx, "'%s': Unrecognized escape sequence '\\%c'.\n", format, c);
 				goto error;
 			}
 		} else if (c == '%') {
@@ -604,12 +608,12 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 				goto one_char;
 			}
 
-			struct bfs_printf *directive = new_directive(NULL);
+			struct bfs_printf *directive = new_directive(ctx, NULL);
 			if (!directive) {
 				goto directive_error;
 			}
 			if (dstrapp(&directive->str, c) != 0) {
-				perror("dstrapp()");
+				bfs_perror(ctx, "dstrapp()");
 				goto directive_error;
 			}
 
@@ -629,11 +633,11 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 				case ' ':
 				case '-':
 					if (strchr(directive->str, c)) {
-						bfs_error(cmdline, "'%s': Duplicate flag '%c'.\n", format, c);
+						bfs_error(ctx, "'%s': Duplicate flag '%c'.\n", format, c);
 						goto directive_error;
 					}
 					if (dstrapp(&directive->str, c) != 0) {
-						perror("dstrapp()");
+						bfs_perror(ctx, "dstrapp()");
 						goto directive_error;
 					}
 					continue;
@@ -645,7 +649,7 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 			// Parse the field width
 			while (c >= '0' && c <= '9') {
 				if (dstrapp(&directive->str, c) != 0) {
-					perror("dstrapp()");
+					bfs_perror(ctx, "dstrapp()");
 					goto directive_error;
 				}
 				c = *++i;
@@ -655,7 +659,7 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 			if (c == '.') {
 				do {
 					if (dstrapp(&directive->str, c) != 0) {
-						perror("dstrapp()");
+						bfs_perror(ctx, "dstrapp()");
 						goto directive_error;
 					}
 					c = *++i;
@@ -685,19 +689,19 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 				directive->fn = bfs_printf_f;
 				break;
 			case 'F':
-				if (!cmdline->mtab) {
-					bfs_error(cmdline, "Couldn't parse the mount table: %s.\n", strerror(cmdline->mtab_error));
+				directive->ptr = bfs_ctx_mtab(ctx);
+				if (!directive->ptr) {
+					bfs_error(ctx, "Couldn't parse the mount table: %m.\n");
 					goto directive_error;
 				}
-				directive->ptr = cmdline->mtab;
 				directive->fn = bfs_printf_F;
 				break;
 			case 'g':
-				if (!cmdline->groups) {
-					bfs_error(cmdline, "Couldn't parse the group table: %s.\n", strerror(cmdline->groups_error));
+				directive->ptr = bfs_ctx_groups(ctx);
+				if (!directive->ptr) {
+					bfs_error(ctx, "Couldn't parse the group table: %m.\n");
 					goto directive_error;
 				}
-				directive->ptr = cmdline->groups;
 				directive->fn = bfs_printf_g;
 				break;
 			case 'G':
@@ -746,11 +750,11 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 				directive->stat_field = BFS_STAT_MTIME;
 				break;
 			case 'u':
-				if (!cmdline->users) {
-					bfs_error(cmdline, "Couldn't parse the user table: %s.\n", strerror(cmdline->users_error));
+				directive->ptr = bfs_ctx_users(ctx);
+				if (!directive->ptr) {
+					bfs_error(ctx, "Couldn't parse the user table: %m.\n");
 					goto directive_error;
 				}
-				directive->ptr = cmdline->users;
 				directive->fn = bfs_printf_u;
 				break;
 			case 'U':
@@ -785,32 +789,32 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 				directive->fn = bfs_printf_strftime;
 				c = *++i;
 				if (!c) {
-					bfs_error(cmdline, "'%s': Incomplete time specifier '%s%c'.\n", format, directive->str, i[-1]);
+					bfs_error(ctx, "'%s': Incomplete time specifier '%s%c'.\n", format, directive->str, i[-1]);
 					goto directive_error;
 				} else if (strchr("%+@aAbBcCdDeFgGhHIjklmMnprRsStTuUVwWxXyYzZ", c)) {
 					directive->c = c;
 				} else {
-					bfs_error(cmdline, "'%s': Unrecognized time specifier '%%%c%c'.\n", format, i[-1], c);
+					bfs_error(ctx, "'%s': Unrecognized time specifier '%%%c%c'.\n", format, i[-1], c);
 					goto directive_error;
 				}
 				break;
 
 			case '\0':
-				bfs_error(cmdline, "'%s': Incomplete format specifier '%s'.\n", format, directive->str);
+				bfs_error(ctx, "'%s': Incomplete format specifier '%s'.\n", format, directive->str);
 				goto directive_error;
 
 			default:
-				bfs_error(cmdline, "'%s': Unrecognized format specifier '%%%c'.\n", format, c);
+				bfs_error(ctx, "'%s': Unrecognized format specifier '%%%c'.\n", format, c);
 				goto directive_error;
 			}
 
 			if (must_be_numeric && strcmp(specifier, "s") == 0) {
-				bfs_error(cmdline, "'%s': Invalid flags '%s' for string format '%%%c'.\n", format, directive->str + 1, c);
+				bfs_error(ctx, "'%s': Invalid flags '%s' for string format '%%%c'.\n", format, directive->str + 1, c);
 				goto directive_error;
 			}
 
 			if (dstrcat(&directive->str, specifier) != 0) {
-				perror("dstrcat()");
+				bfs_perror(ctx, "dstrcat()");
 				goto directive_error;
 			}
 
@@ -818,7 +822,7 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 			tail = append_directive(tail, directive);
 
 			if (!literal) {
-				literal = new_directive(bfs_printf_literal);
+				literal = new_directive(ctx, bfs_printf_literal);
 				if (!literal) {
 					goto error;
 				}
@@ -833,7 +837,7 @@ struct bfs_printf *parse_bfs_printf(const char *format, struct cmdline *cmdline)
 
 	one_char:
 		if (dstrapp(&literal->str, c) != 0) {
-			perror("dstrapp()");
+			bfs_perror(ctx, "dstrapp()");
 			goto error;
 		}
 	}
@@ -849,14 +853,14 @@ done:
 
 error:
 	free_directive(literal);
-	free_bfs_printf(head);
+	bfs_printf_free(head);
 	return NULL;
 }
 
-int bfs_printf(FILE *file, const struct bfs_printf *command, const struct BFTW *ftwbuf) {
+int bfs_printf(FILE *file, const struct bfs_printf *format, const struct BFTW *ftwbuf) {
 	int ret = 0, error = 0;
 
-	for (const struct bfs_printf *directive = command; directive; directive = directive->next) {
+	for (const struct bfs_printf *directive = format; directive; directive = directive->next) {
 		if (directive->fn(file, directive, ftwbuf) < 0) {
 			ret = -1;
 			error = errno;
@@ -867,10 +871,10 @@ int bfs_printf(FILE *file, const struct bfs_printf *command, const struct BFTW *
 	return ret;
 }
 
-void free_bfs_printf(struct bfs_printf *command) {
-	while (command) {
-		struct bfs_printf *next = command->next;
-		free_directive(command);
-		command = next;
+void bfs_printf_free(struct bfs_printf *format) {
+	while (format) {
+		struct bfs_printf *next = format->next;
+		free_directive(format);
+		format = next;
 	}
 }
