@@ -166,6 +166,10 @@ posix_tests=(
 
     test_weird_names
 
+    test_incomplete
+    test_missing_paren
+    test_extra_paren
+
     # Flags
 
     test_H
@@ -196,6 +200,7 @@ posix_tests=(
 
     test_group_name
     test_group_id
+    test_group_nogroup
 
     test_links
     test_links_plus
@@ -209,6 +214,12 @@ posix_tests=(
     test_newer
     test_newer_link
 
+    test_nogroup
+    test_nogroup_ulimit
+
+    test_nouser
+    test_nouser_ulimit
+
     test_ok_stdin
     test_ok_plus_semicolon
 
@@ -224,6 +235,8 @@ posix_tests=(
     test_perm_symbolic_minus
     test_perm_leading_plus_symbolic_minus
     test_permcopy
+    test_perm_setid
+    test_perm_sticky
 
     test_prune
     test_prune_or_print
@@ -238,6 +251,7 @@ posix_tests=(
 
     test_user_name
     test_user_id
+    test_user_nouser
 
     # Closed file descriptors
     test_closed_stdin
@@ -253,6 +267,8 @@ posix_tests=(
     test_de_morgan_not
     test_de_morgan_and
     test_de_morgan_or
+    test_data_flow_group
+    test_data_flow_user
     test_data_flow_type
     test_data_flow_and_swap
     test_data_flow_or_swap
@@ -329,6 +345,9 @@ bsd_tests=(
     test_lname
     test_L_lname
 
+    test_ls
+    test_L_ls
+
     test_maxdepth
 
     test_mindepth
@@ -347,12 +366,6 @@ bsd_tests=(
     test_newermt
     test_newermt_epoch_minus_one
 
-    test_nogroup
-    test_nogroup_ulimit
-
-    test_nouser
-    test_nouser_ulimit
-
     test_ok_stdin
     test_ok_closed_stdin
 
@@ -452,12 +465,18 @@ gnu_tests=(
 
     test_false
 
+    test_fls
+
     test_follow
 
     test_fprint
     test_fprint_duplicate
     test_fprint_error
 
+    test_fprint0
+
+    test_fprintf
+
     test_fstype
 
     test_gid
@@ -484,6 +503,9 @@ gnu_tests=(
     test_lname
     test_L_lname
 
+    test_ls
+    test_L_ls
+
     test_maxdepth
 
     test_mindepth
@@ -497,12 +519,6 @@ gnu_tests=(
     test_newermt
     test_newermt_epoch_minus_one
 
-    test_nogroup
-    test_nogroup_ulimit
-
-    test_nouser
-    test_nouser_ulimit
-
     test_ok_closed_stdin
     test_ok_nothing
 
@@ -517,6 +533,8 @@ gnu_tests=(
 
     test_print_error
 
+    test_print0
+
     test_printf
     test_printf_empty
     test_printf_slash
@@ -532,6 +550,7 @@ gnu_tests=(
     test_printf_Y_error
     test_printf_H
     test_printf_u_g_ulimit
+    test_printf_l_nonlink
 
     test_quit
     test_quit_child
@@ -544,6 +563,7 @@ gnu_tests=(
 
     test_regex
     test_regex_parens
+    test_regex_error
 
     test_regextype_posix_basic
     test_regextype_posix_extended
@@ -583,6 +603,7 @@ gnu_tests=(
     test_and_false_or_true
     test_comma_redundant_true
     test_comma_redundant_false
+    test_data_flow_sparse
 )
 
 bfs_tests=(
@@ -593,12 +614,31 @@ bfs_tests=(
     test_expr_flag_path
     test_expr_path_flag
 
+    test_typo
+
     # Flags
 
+    test_D_multi
+    test_D_all
+
+    test_O0
+    test_O1
+    test_O2
+    test_O3
+    test_Ofast
+
     test_S_bfs
     test_S_dfs
     test_S_ids
 
+    # Special forms
+
+    test_exclude_name
+    test_exclude_depth
+    test_exclude_mindepth
+    test_exclude_print
+    test_exclude_exclude
+
     # Primaries
 
     test_color
@@ -642,7 +682,13 @@ bfs_tests=(
 
     test_hidden
 
+    test_newerma_nonexistent
+    test_newermt_invalid
+    test_newermq
+    test_newerqm
+
     test_nohidden
+    test_nohidden_depth
 
     test_perm_symbolic_trailing_comma
     test_perm_symbolic_double_comma
@@ -650,6 +696,12 @@ bfs_tests=(
     test_perm_leading_plus_symbolic
 
     test_printf_w
+    test_printf_incomplete_escape
+    test_printf_invalid_escape
+    test_printf_incomplete_format
+    test_printf_invalid_format
+    test_printf_duplicate_flag
+    test_printf_must_be_numeric
 
     test_type_multi
 
@@ -659,8 +711,14 @@ bfs_tests=(
     test_L_unique_loops
     test_L_unique_depth
 
+    test_version
+
     test_xtype_multi
+
+    # Optimizer tests
+    test_data_flow_hidden
     test_xtype_reorder
+    test_xtype_depth
 
     # PATH_MAX handling
     test_deep_strict
@@ -686,12 +744,18 @@ case "$UNAME" in
         bsd_tests+=(
             test_xattr
             test_L_xattr
+
+            test_xattrname
+            test_L_xattrname
         )
         ;;
     *)
         sudo_tests+=(
             test_xattr
             test_L_xattr
+
+            test_xattrname
+            test_L_xattrname
         )
         ;;
 esac
@@ -939,6 +1003,15 @@ function invoke_bfs() {
     $BFS "$@"
 }
 
+# Silence stderr unless --verbose is set
+function quiet() {
+    if [ "$VERBOSE" ]; then
+        "$@"
+    else
+        "$@" 2>/dev/null
+    fi
+}
+
 # Return value when bfs fails
 EX_BFS=10
 # Return value when a difference is detected
@@ -951,11 +1024,18 @@ function bfs_diff() (
     # substitution, even with low ulimit -n
     exec 3>&-
 
-    local EXPECTED="$TESTS/${FUNCNAME[1]}.out"
+    local CALLER
+    for CALLER in "${FUNCNAME[@]}"; do
+        if [[ $CALLER == test_* ]]; then
+            break
+        fi
+    done
+
+    local EXPECTED="$TESTS/$CALLER.out"
     if [ "$UPDATE" ]; then
         local ACTUAL="$EXPECTED"
     else
-        local ACTUAL="$TMP/${FUNCNAME[1]}.out"
+        local ACTUAL="$TMP/$CALLER.out"
     fi
 
     $BFS "$@" | bfs_sort >"$ACTUAL"
@@ -1054,7 +1134,7 @@ function test_depth_error() {
     touchp scratch/foo/bar
     chmod -r scratch/foo
 
-    bfs_diff scratch -depth 2>/dev/null
+    quiet bfs_diff scratch -depth
     local ret=$?
 
     chmod +r scratch/foo
@@ -1239,12 +1319,12 @@ function test_L_loops() {
 }
 
 function test_L_loops_continue() {
-    bfs_diff -L loops 2>/dev/null
+    quiet bfs_diff -L loops
     [ $? -eq $EX_BFS ]
 }
 
 function test_X() {
-    bfs_diff -X weirdnames 2>/dev/null
+    quiet bfs_diff -X weirdnames
     [ $? -eq $EX_BFS ]
 }
 
@@ -1323,6 +1403,11 @@ function test_xtype_reorder() {
     invoke_bfs loops -links 100 -xtype l
 }
 
+function test_xtype_depth() {
+    # Make sure -xtype is considered side-effecting for facts_when_impure
+    ! quiet invoke_bfs loops -xtype l -depth 100
+}
+
 function test_iname() {
     bfs_diff basic -iname '*F*'
 }
@@ -1355,6 +1440,11 @@ function test_user_id() {
     bfs_diff basic -user "$(id -u)"
 }
 
+function test_user_nouser() {
+    # Regression test: this was wrongly optimized to -false
+    bfs_diff basic -user "$(id -u)" \! -nouser
+}
+
 function test_group_name() {
     bfs_diff basic -group "$(id -gn)"
 }
@@ -1363,6 +1453,11 @@ function test_group_id() {
     bfs_diff basic -group "$(id -g)"
 }
 
+function test_group_nogroup() {
+    # Regression test: this was wrongly optimized to -false
+    bfs_diff basic -group "$(id -g)" \! -nogroup
+}
+
 function test_daystart() {
     bfs_diff basic -daystart -mtime 0
 }
@@ -1383,6 +1478,22 @@ function test_newermt_epoch_minus_one() {
     bfs_diff times -newermt 1969-12-31T23:59:59Z
 }
 
+function test_newermt_invalid() {
+    ! quiet invoke_bfs times -newermt not_a_date_time
+}
+
+function test_newerma_nonexistent() {
+    ! quiet invoke_bfs times -newerma basic/nonexistent
+}
+
+function test_newermq() {
+    ! quiet invoke_bfs times -newermq times/a
+}
+
+function test_newerqm() {
+    ! quiet invoke_bfs times -newerqm times/a
+}
+
 function test_size() {
     bfs_diff basic -type f -size 0
 }
@@ -1405,7 +1516,7 @@ function test_exec() {
 
 function test_exec_nothing() {
     # Regression test: don't segfault on missing command
-    ! invoke_bfs basic -exec \; 2>/dev/null
+    ! quiet invoke_bfs basic -exec \;
 }
 
 function test_exec_plus() {
@@ -1546,7 +1657,7 @@ function test_ignore_readdir_race() {
 
 function test_ignore_readdir_race_root() {
     # Make sure -ignore_readdir_race doesn't suppress ENOENT at the root
-    ! invoke_bfs basic/nonexistent -ignore_readdir_race 2>/dev/null
+    ! quiet invoke_bfs basic/nonexistent -ignore_readdir_race
 }
 
 function test_ignore_readdir_race_notdir() {
@@ -1618,15 +1729,15 @@ function test_perm_symbolic_slash() {
 }
 
 function test_perm_symbolic_trailing_comma() {
-    ! invoke_bfs perms -perm a+r, 2>/dev/null
+    ! quiet invoke_bfs perms -perm a+r,
 }
 
 function test_perm_symbolic_double_comma() {
-    ! invoke_bfs perms -perm a+r,,u+w 2>/dev/null
+    ! quiet invoke_bfs perms -perm a+r,,u+w
 }
 
 function test_perm_symbolic_missing_action() {
-    ! invoke_bfs perms -perm a 2>/dev/null
+    ! quiet invoke_bfs perms -perm a
 }
 
 function test_perm_leading_plus_symbolic() {
@@ -1645,6 +1756,14 @@ function test_permcopy() {
     bfs_diff perms -perm u+rw,g+u-w,o=g
 }
 
+function test_perm_setid() {
+    bfs_diff rainbow -perm /ug+s
+}
+
+function test_perm_sticky() {
+    bfs_diff rainbow -perm /ug+t
+}
+
 function test_prune() {
     bfs_diff basic -name foo -prune
 }
@@ -1659,26 +1778,26 @@ function test_not_prune() {
 
 function test_ok_nothing() {
     # Regression test: don't segfault on missing command
-    ! invoke_bfs basic -ok \; 2>/dev/null
+    ! quiet invoke_bfs basic -ok \;
 }
 
 function test_ok_stdin() {
     # -ok should *not* close stdin
     # See https://savannah.gnu.org/bugs/?24561
-    yes | bfs_diff basic -ok bash -c 'printf "%s? " "$1" && head -n1' bash '{}' \; 2>/dev/null
+    yes | quiet bfs_diff basic -ok bash -c 'printf "%s? " "$1" && head -n1' bash '{}' \;
 }
 
 function test_okdir_stdin() {
     # -okdir should *not* close stdin
-    yes | bfs_diff basic -okdir bash -c 'printf "%s? " "$1" && head -n1' bash '{}' \; 2>/dev/null
+    yes | quiet bfs_diff basic -okdir bash -c 'printf "%s? " "$1" && head -n1' bash '{}' \;
 }
 
 function test_ok_plus_semicolon() {
-    yes | bfs_diff basic -ok echo '{}' + \; 2>/dev/null
+    yes | quiet bfs_diff basic -ok echo '{}' + \;
 }
 
 function test_okdir_plus_semicolon() {
-    yes | bfs_diff basic -okdir echo '{}' + \; 2>/dev/null
+    yes | quiet bfs_diff basic -okdir echo '{}' + \;
 }
 
 function test_delete() {
@@ -1725,6 +1844,10 @@ function test_regex_parens() {
     bfs_diff . -regex '\./\((\)'
 }
 
+function test_regex_error() {
+    ! quiet invoke_bfs basic -regex '['
+}
+
 function test_E() {
     cd weirdnames
     bfs_diff -E . -regex '\./(\()'
@@ -1771,6 +1894,10 @@ function test_nohidden() {
     bfs_diff weirdnames -nohidden
 }
 
+function test_nohidden_depth() {
+    bfs_diff weirdnames -depth -nohidden
+}
+
 function test_depth_n() {
     bfs_diff basic -depth 2
 }
@@ -1875,6 +2002,18 @@ function test_nouser_ulimit() {
     bfs_diff deep -nouser
 }
 
+function test_ls() {
+    invoke_bfs rainbow -ls >scratch/test_ls.out
+}
+
+function test_L_ls() {
+    invoke_bfs -L rainbow -ls >scratch/test_L_ls.out
+}
+
+function test_fls() {
+    invoke_bfs rainbow -fls scratch/test_fls.out
+}
+
 function test_printf() {
     bfs_diff basic -printf '%%p(%p) %%d(%d) %%f(%f) %%h(%h) %%H(%H) %%P(%P) %%m(%m) %%M(%M) %%y(%y)\n'
 }
@@ -1947,7 +2086,7 @@ function test_printf_Y_error() {
     chmod -x scratch/foo
     ln -s foo/bar scratch/bar
 
-    bfs_diff scratch -printf '(%p) (%l) %y %Y\n' 2>/dev/null
+    quiet bfs_diff scratch -printf '(%p) (%l) %y %Y\n'
     local ret=$?
 
     chmod +x scratch/foo
@@ -1966,6 +2105,45 @@ function test_printf_u_g_ulimit() {
     [ "$(invoke_bfs deep -printf '%u %g\n' | uniq)" = "$(id -un) $(id -gn)" ]
 }
 
+function test_printf_l_nonlink() {
+    bfs_diff links -printf '| %24p -> %-24l |\n'
+}
+
+function test_printf_incomplete_escape() {
+    ! quiet invoke_bfs basic -printf '\'
+}
+
+function test_printf_invalid_escape() {
+    ! quiet invoke_bfs basic -printf '\!'
+}
+
+function test_printf_incomplete_format() {
+    ! quiet invoke_bfs basic -printf '%'
+}
+
+function test_printf_invalid_format() {
+    ! quiet invoke_bfs basic -printf '%!'
+}
+
+function test_printf_duplicate_flag() {
+    ! quiet invoke_bfs basic -printf '%--p'
+}
+
+function test_printf_must_be_numeric() {
+    ! quiet invoke_bfs basic -printf '%+p'
+}
+
+function test_fprintf() {
+    invoke_bfs basic -fprintf scratch/test_fprintf.out '%%p(%p) %%d(%d) %%f(%f) %%h(%h) %%H(%H) %%P(%P) %%m(%m) %%M(%M) %%y(%y)\n'
+    sort -o scratch/test_fprintf.out scratch/test_fprintf.out
+
+    if [ "$UPDATE" ]; then
+        cp scratch/test_fprintf.out "$TESTS/test_fprintf.out"
+    else
+        diff -u scratch/test_fprintf.out "$TESTS/test_fprintf.out"
+    fi
+}
+
 function test_fstype() {
     fstype="$(invoke_bfs basic -maxdepth 0 -printf '%F\n')"
     bfs_diff basic -fstype "$fstype"
@@ -2031,6 +2209,18 @@ function test_precedence() {
     bfs_diff basic \( -name foo -type d -o -name bar -a -type f \) -print , \! -empty -type f -print
 }
 
+function test_incomplete() {
+    ! quiet invoke_bfs basic \(
+}
+
+function test_missing_paren() {
+    ! quiet invoke_bfs basic \( -print
+}
+
+function test_extra_paren() {
+    ! quiet invoke_bfs basic -print \)
+}
+
 function test_color() {
     LS_COLORS= bfs_diff rainbow -color
 }
@@ -2151,11 +2341,11 @@ function test_color_ext0() {
 }
 
 function test_color_ext_override() {
-    LS_COLORS="*.tar.gz=01;31:*.tar=01;32:*.gz=01;33:" bfs_diff rainbow -color
+    LS_COLORS="*.tar.gz=01;31:*.TAR=01;32:*.gz=01;33:" bfs_diff rainbow -color
 }
 
 function test_color_ext_underride() {
-    LS_COLORS="*.gz=01;33:*.tar=01;32:*.tar.gz=01;31:" bfs_diff rainbow -color
+    LS_COLORS="*.gz=01;33:*.TAR=01;32:*.tar.gz=01;31:" bfs_diff rainbow -color
 }
 
 function test_color_missing_colon() {
@@ -2289,6 +2479,22 @@ function test_data_flow_depth() {
     bfs_diff basic -depth +1 -depth -4
 }
 
+function test_data_flow_group() {
+    bfs_diff basic \( -group "$(id -g)" -nogroup \) -o \( -group "$(id -g)" -o -nogroup \)
+}
+
+function test_data_flow_user() {
+    bfs_diff basic \( -user "$(id -u)" -nouser \) -o \( -user "$(id -u)" -o -nouser \)
+}
+
+function test_data_flow_hidden() {
+    bfs_diff basic \( -hidden -not -hidden \) -o \( -hidden -o -not -hidden \)
+}
+
+function test_data_flow_sparse() {
+    bfs_diff basic \( -sparse -not -sparse \) -o \( -sparse -o -not -sparse \)
+}
+
 function test_data_flow_type() {
     bfs_diff basic \! \( -type f -o \! -type f \)
 }
@@ -2303,13 +2509,33 @@ function test_data_flow_or_swap() {
 
 function test_print_error() {
     if [ -e /dev/full ]; then
-        ! invoke_bfs basic -maxdepth 0 >/dev/full 2>/dev/null
+        ! quiet invoke_bfs basic -maxdepth 0 >/dev/full
     fi
 }
 
 function test_fprint_error() {
     if [ -e /dev/full ]; then
-        ! invoke_bfs basic -maxdepth 0 -fprint /dev/full 2>/dev/null
+        ! quiet invoke_bfs basic -maxdepth 0 -fprint /dev/full
+    fi
+}
+
+function test_print0() {
+    invoke_bfs basic/a basic/b -print0 >scratch/test_print0.out
+
+    if [ "$UPDATE" ]; then
+        cp scratch/test_print0.out "$TESTS/test_print0.out"
+    else
+        cmp -s scratch/test_print0.out "$TESTS/test_print0.out"
+    fi
+}
+
+function test_fprint0() {
+    invoke_bfs basic/a basic/b -fprint0 scratch/test_fprint0.out
+
+    if [ "$UPDATE" ]; then
+        cp scratch/test_fprint0.out "$TESTS/test_fprint0.out"
+    else
+        cmp -s scratch/test_fprint0.out "$TESTS/test_fprint0.out"
     fi
 }
 
@@ -2318,15 +2544,15 @@ function test_closed_stdin() {
 }
 
 function test_ok_closed_stdin() {
-    bfs_diff basic -ok echo \; <&- 2>/dev/null
+    quiet bfs_diff basic -ok echo \; <&-
 }
 
 function test_okdir_closed_stdin() {
-    bfs_diff basic -okdir echo {} \; <&- 2>/dev/null
+    quiet bfs_diff basic -okdir echo {} \; <&-
 }
 
 function test_closed_stdout() {
-    ! invoke_bfs basic >&- 2>/dev/null
+    ! quiet invoke_bfs basic >&-
 }
 
 function test_closed_stderr() {
@@ -2479,7 +2705,7 @@ function set_acl() {
 function test_acl() {
     rm -rf scratch/*
 
-    invoke_bfs scratch -quit -acl 2>/dev/null || return 0
+    quiet invoke_bfs scratch -quit -acl || return 0
 
     touch scratch/{normal,acl}
     set_acl scratch/acl || return 0
@@ -2491,7 +2717,7 @@ function test_acl() {
 function test_L_acl() {
     rm -rf scratch/*
 
-    invoke_bfs scratch -quit -acl 2>/dev/null || return 0
+    quiet invoke_bfs scratch -quit -acl || return 0
 
     touch scratch/{normal,acl}
     set_acl scratch/acl || return 0
@@ -2503,7 +2729,7 @@ function test_L_acl() {
 function test_capable() {
     rm -rf scratch/*
 
-    if ! invoke_bfs scratch -quit -capable 2>/dev/null; then
+    if ! quiet invoke_bfs scratch -quit -capable; then
         return 0
     fi
 
@@ -2517,7 +2743,7 @@ function test_capable() {
 function test_L_capable() {
     rm -rf scratch/*
 
-    if ! invoke_bfs scratch -quit -capable 2>/dev/null; then
+    if ! quiet invoke_bfs scratch -quit -capable; then
         return 0
     fi
 
@@ -2528,55 +2754,86 @@ function test_L_capable() {
     bfs_diff -L scratch -capable
 }
 
-function set_xattr() {
+function make_xattrs() {
+    rm -rf scratch/*
+
+    touch scratch/{normal,xattr,xattr_2}
+    ln -s xattr scratch/link
+    ln -s normal scratch/xattr_link
+
     case "$UNAME" in
         Darwin)
-            xattr -w bfs_test true "$1"
-            xattr -s -w bfs_test true "$2"
+            xattr -w bfs_test true scratch/xattr
+            xattr -w bfs_test_2 true scratch/xattr_2
+            xattr -s -w bfs_test true scratch/xattr_link
             ;;
         FreeBSD)
-            setextattr user bfs_test true "$1"
-            setextattr -h user bfs_test true "$2"
+            setextattr user bfs_test true scratch/xattr
+            setextattr user bfs_test_2 true scratch/xattr_2
+            setextattr -h user bfs_test true scratch/xattr_link
             ;;
         *)
             # Linux tmpfs doesn't support the user.* namespace, so we use the security.*
             # namespace, which is writable by root and readable by others
-            sudo setfattr -n security.bfs_test "$1"
-            sudo setfattr -h -n security.bfs_test "$2"
+            sudo setfattr -n security.bfs_test scratch/xattr
+            sudo setfattr -n security.bfs_test_2 scratch/xattr_2
+            sudo setfattr -h -n security.bfs_test scratch/xattr_link
             ;;
     esac
 }
 
 function test_xattr() {
-    rm -rf scratch/*
-
-    if ! invoke_bfs scratch -quit -xattr 2>/dev/null; then
+    if ! quiet invoke_bfs scratch -quit -xattr; then
         return 0
     fi
 
-    touch scratch/{normal,xattr}
-    ln -s xattr scratch/link
-    ln -s normal scratch/xattr_link
-    set_xattr scratch/xattr scratch/xattr_link
-
+    make_xattrs
     bfs_diff scratch -xattr
 }
 
 function test_L_xattr() {
-    rm -rf scratch/*
-
-    if ! invoke_bfs scratch -quit -xattr 2>/dev/null; then
+    if ! quiet invoke_bfs scratch -quit -xattr; then
         return 0
     fi
 
-    touch scratch/{normal,xattr}
-    ln -s xattr scratch/link
-    ln -s normal scratch/xattr_link
-    set_xattr scratch/xattr scratch/xattr_link
-
+    make_xattrs
     bfs_diff -L scratch -xattr
 }
 
+function test_xattrname() {
+    if ! quiet invoke_bfs scratch -quit -xattr; then
+        return 0
+    fi
+
+    make_xattrs
+
+    case "$UNAME" in
+        Darwin|FreeBSD)
+            bfs_diff scratch -xattrname bfs_test
+            ;;
+        *)
+            bfs_diff scratch -xattrname security.bfs_test
+            ;;
+    esac
+}
+
+function test_L_xattrname() {
+    if ! quiet invoke_bfs scratch -quit -xattr; then
+        return 0
+    fi
+
+    make_xattrs
+
+    case "$UNAME" in
+        Darwin|FreeBSD)
+            bfs_diff -L scratch -xattrname bfs_test
+            ;;
+        *)
+            bfs_diff -L scratch -xattrname security.bfs_test
+            ;;
+    esac
+}
+
 function test_help() {
     invoke_bfs -help | grep -E '\{...?\}' && return 1
     invoke_bfs -D help | grep -E '\{...?\}' && return 1
@@ -2586,6 +2843,42 @@ function test_help() {
     return 0
 }
 
+function test_version() {
+    invoke_bfs -version >/dev/null
+}
+
+function test_typo() {
+    invoke_bfs -dikkiq 2>&1 | grep follow >/dev/null
+}
+
+function test_D_multi() {
+    quiet bfs_diff -D opt,tree,unknown basic
+}
+
+function test_D_all() {
+    quiet bfs_diff -D all basic
+}
+
+function test_O0() {
+    bfs_diff -O0 basic -not \( -type f -not -type f \)
+}
+
+function test_O1() {
+    bfs_diff -O1 basic -not \( -type f -not -type f \)
+}
+
+function test_O2() {
+    bfs_diff -O2 basic -not \( -type f -not -type f \)
+}
+
+function test_O3() {
+    bfs_diff -O3 basic -not \( -type f -not -type f \)
+}
+
+function test_Ofast() {
+    bfs_diff -Ofast basic -not \( -xtype f -not -xtype f \)
+}
+
 function test_S() {
     invoke_bfs -S "$1" -s basic >"$TMP/test_S_$1.out"
 
@@ -2608,6 +2901,26 @@ function test_S_ids() {
     test_S ids
 }
 
+function test_exclude_name() {
+    bfs_diff basic -exclude -name foo
+}
+
+function test_exclude_depth() {
+    bfs_diff basic -depth -exclude -name foo
+}
+
+function test_exclude_mindepth() {
+    bfs_diff basic -mindepth 3 -exclude -name foo
+}
+
+function test_exclude_print() {
+    ! quiet invoke_bfs basic -exclude -print
+}
+
+function test_exclude_exclude() {
+    ! quiet invoke_bfs basic -exclude -exclude -name foo
+}
+
 
 BOL=
 EOL='\n'
