@@ -15,7 +15,6 @@
  ****************************************************************************/
 
 #include "util.h"
-#include "bftw.h"
 #include "dstring.h"
 #include <errno.h>
 #include <fcntl.h>
@@ -59,10 +58,15 @@ int xreaddir(DIR *dir, struct dirent **de) {
 }
 
 char *xreadlinkat(int fd, const char *path, size_t size) {
-	++size; // NUL-terminator
 	ssize_t len;
 	char *name = NULL;
 
+	if (size == 0) {
+		size = 64;
+	} else {
+		++size; // NUL terminator
+	}
+
 	while (true) {
 		char *new_name = realloc(name, size);
 		if (!new_name) {
@@ -73,7 +77,7 @@ char *xreadlinkat(int fd, const char *path, size_t size) {
 		len = readlinkat(fd, path, name, size);
 		if (len < 0) {
 			goto error;
-		} else if (len >= size) {
+		} else if ((size_t)len >= size) {
 			size *= 2;
 		} else {
 			break;
@@ -88,34 +92,6 @@ error:
 	return NULL;
 }
 
-bool isopen(int fd) {
-	return fcntl(fd, F_GETFD) >= 0 || errno != EBADF;
-}
-
-int redirect(int fd, const char *path, int flags, ...) {
-	mode_t mode = 0;
-	if (flags & O_CREAT) {
-		va_list args;
-		va_start(args, flags);
-
-		// Use int rather than mode_t, because va_arg must receive a
-		// fully-promoted type
-		mode = va_arg(args, int);
-
-		va_end(args);
-	}
-
-	int ret = open(path, flags, mode);
-
-	if (ret >= 0 && ret != fd) {
-		int orig = ret;
-		ret = dup2(orig, fd);
-		close(orig);
-	}
-
-	return ret;
-}
-
 int dup_cloexec(int fd) {
 #ifdef F_DUPFD_CLOEXEC
 	return fcntl(fd, F_DUPFD_CLOEXEC, 0);
@@ -163,44 +139,44 @@ char *xregerror(int err, const regex_t *regex) {
 	return str;
 }
 
-void format_mode(mode_t mode, char str[11]) {
+/** Get the single character describing the given file type. */
+static char type_char(mode_t mode) {
+	switch (mode & S_IFMT) {
+	case S_IFREG:
+		return '-';
+	case S_IFBLK:
+		return 'b';
+	case S_IFCHR:
+		return 'c';
+	case S_IFDIR:
+		return 'd';
+	case S_IFLNK:
+		return 'l';
+	case S_IFIFO:
+		return 'p';
+	case S_IFSOCK:
+		return 's';
+#ifdef S_IFDOOR
+	case S_IFDOOR:
+		return 'D';
+#endif
+#ifdef S_IFPORT
+	case S_IFPORT:
+		return 'P';
+#endif
+#ifdef S_IFWHT
+	case S_IFWHT:
+		return 'w';
+#endif
+	}
+
+	return '?';
+}
+
+void xstrmode(mode_t mode, char str[11]) {
 	strcpy(str, "----------");
 
-	switch (bftw_mode_typeflag(mode)) {
-	case BFTW_REG:
-		break;
-	case BFTW_BLK:
-		str[0] = 'b';
-		break;
-	case BFTW_CHR:
-		str[0] = 'c';
-		break;
-	case BFTW_DIR:
-		str[0] = 'd';
-		break;
-	case BFTW_DOOR:
-		str[0] = 'D';
-		break;
-	case BFTW_LNK:
-		str[0] = 'l';
-		break;
-	case BFTW_FIFO:
-		str[0] = 'p';
-		break;
-	case BFTW_PORT:
-		str[0] = 'P';
-		break;
-	case BFTW_SOCK:
-		str[0] = 's';
-		break;
-	case BFTW_WHT:
-		str[0] = 'w';
-		break;
-	case BFTW_UNKNOWN:
-	case BFTW_ERROR:
-		str[0] = '?';
-		break;
-	}
+	str[0] = type_char(mode);
 
 	if (mode & 00400) {
 		str[1] = 'r';
